--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

--------------------
-- Initialization --
--------------------

function InitAntiPosLag()
	if AntiPosLag then
		-- Save the original API
		PreAntiLag_Move = Move

		-- Set the new move command
		Move = AntiLag_Move

		-- Set the check command
		CheckPosLag = AI_CheckPosLag

		-- Set the command to bypass anti-poslag
		BypassMove = AntiLag_Bypass
	else
		-- Set a check command that returns immediately
		CheckPosLag = FalseReturn

		-- Set a command to bypass anti-poslag
		BypassMove = Move
	end
end

------------------
-- Move Command --
------------------

function AntiLag_Bypass(actor,x,y)
	AntiLag_Move(actor,x,y,true)
end

function AntiLag_Move(actor,x,y,bypass)
	-- Check the actor
	if actor ~= HomunID then
		-- Can't do anything
		return
	end

	-- Check if we're supposed to bypass
	if not bypass then
		-- Get the tick count
		local ticks = GetTick()
	
		-- Check if its the same as our current destination
		if x == PosLag.move.x and y == PosLag.move.y then
			-- Check if we still are trying a previous move
			if PosLag.tries.left > 0 then
				-- Don't do anything
				return
			end

			-- Check if we're already at the location
			local homux,homuy = GetV(V_POSITION,HomunID)
			if homux == x and homuy == y then
				-- Check if we're moving
				if HomunMotion == MOTION_MOVE then
					-- Stop moving
					return PreAntiLag_Move(HomunID,x,y)
				end

				-- Otherwise just return
				return
			end
		else
			-- Set the PosLag destination
			PosLag.move.x,PosLag.move.y,PosLag.move.ticks = x,y,ticks
		end
	
		-- We're trying
		PosLag.tries.trying = true
	
		-- Set the PosLag current location (and first time there)
		PosLag.check.x,PosLag.check.y = GetV(V_POSITION,HomunID)
		PosLag.check.ticks = ticks
	
		-- Set the number of tries
		PosLag.tries.left,PosLag.tries.timeout = PosLagTries+1,ticks + PosLagDelay
	else
		-- Bypass anti-position lag
		PosLag.tries.left,PosLag.tries.timeout = 0,0
		PosLag.tries.trying = false
	end

	-- Do the actual move
	return PreAntiLag_Move(HomunID,x,y)
end

----------------------------
-- Check for Position Lag --
----------------------------

function AI_CheckPosLag()
	-- Get our current location
	local x,y = GetV(V_POSITION,HomunID)

	-- Get the current number of ticks
	local ticks = GetTick()

	-- Check to see if we have any tries left
	if PosLag.tries.left < 1 then
		-- No more tries

		-- We're no longer trying
		PosLag.tries.trying = false

		-- We're not busy
		return false
	end

	-- Check to see if we're moving
	if GetV(V_MOTION,HomunID) == MOTION_MOVE then
		-- Waiting until we're not moving
		return true
	end

	-- Check if we've moved since our last position
	if PosLag.check.x ~= x and PosLag.check.y ~= y then
		-- Update the current position
		PosLag.check.x,PosLag.check.y = x,y
		PosLag.check.ticks = ticks

		-- Reset the tries and timeout
		PosLag.tries.left = PosLagTries
		PosLag.tries.timeout = ticks + PosLagDelay
	end

	-- Check if we're at our destination
	if PPDistance(x,y,PosLag.move.x,PosLag.move.y) <= 1 then
		-- Check if we should log it
		if PosLag.tries.msg.x ~= PosLag.check.x and
			PosLag.tries.msg.y ~= PosLag.check.y then
			-- Log it
			Log(5,"Arrived at (%d,%d) after %4.3f seconds.",
				PosLag.move.x,PosLag.move.y,(ticks - PosLag.move.ticks) / 1000)
	
			-- Set the message X and Y
			PosLag.tries.msg.x,PosLag.tries.msg.y = PosLag.check.x,PosLag.check.y
		end


		-- Remove tries and return that we're not busy
		PosLag.tries.left = 0
		PosLag.tries.timeout = 0

		-- Send the final move, just to be sure
		PreAntiLag_Move(HomunID,PosLag.move.x,PosLag.move.y)

		-- We're not busy
		return false
	end

	-- Check to see if the timeout is up
	if PosLag.tries.timeout <= ticks and PosLag.tries.timeout ~= 0 then
		-- Check if its time to log it
		if PosLag.tries.msg.x ~= PosLag.check.x and
			PosLag.tries.msg.y ~= PosLag.check.y then

			-- Log it
			Log(2,"Trying to defeat position lag when moving from (%d,%d) to (%d,%d).",
				PosLag.check.x,PosLag.check.y,PosLag.move.x,PosLag.move.y)

			-- Set the message X and Y
			PosLag.tries.msg.x,PosLag.tries.msg.y = PosLag.check.x,PosLag.check.y
		end

		-- Reduce the tries left by 1
		PosLag.tries.left = PosLag.tries.left - 1

		-- Check if there are any tries left
		if PosLag.tries.left < 1 then
			-- Turn off
			PosLag.tries.trying = false
			PosLag.tries.timeout = 0

			-- Not busy
			return false
		end

		-- Reset the timeout
		PosLag.tries.timeout = ticks + PosLagDelay
	else
		-- Still waiting for the timeout, so we're busy
		return true
	end

	-- Grab the type of anti-position lag we're going to try
	local type = math.mod(PosLag.tries.left,2)

	-- Determine where to try our move
	local destx,desty
	if type == 0 then
		-- Try moving to ourself
		destx,desty = x,y
	else --if type == 1 then
		-- Try moving to a spot nearby our destination
		destx,desty = PosLag.move.x,PosLag.move.y

		-- Get a random percent for X
		local percent = math.random(1,100)

		-- Check if its in the upper 50%
		if percent > 50 then
			-- Check if its in the upper 25%
			if percent > 75 then
				-- Decrement X
				destx = destx - 1
			else
				-- Increment X
				destx = destx + 1
			end
		else
			-- Check if its in the lower 25%
			if percent <= 25 then
				-- Decrement Y
				desty = desty - 1
			else
				-- Increment Y
				desty = desty + 1
			end
		end
	end

	-- Make the move
	PreAntiLag_Move(HomunID,destx,desty)

	-- Since we're trying to defeat position lag, we're busy
	return true
end