--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]


--------------------------------------
-- Initialize Auto-Alchemist Skills --
--------------------------------------

function InitOwnerSkills()
	-- Check if the user wants to use auto Homunculus Resurrection
	if AutoRes then
		-- Set the HomuResuCheck function
		HomuResuCheck = AI_HomuResurrect

		-- Add a function that calls HomuResuCheck (so HomuResuCheck can be hooked)
		AddAlchSkill(10,function(castAllowed)
			-- Call HomuResuCheck
			return HomuResuCheck(castAllowed)
		end)
	end

	-- Check if the user wants to cheat. I mean auto-aid-potion. (I'm kidding!)
	if AutoAidPot then
		-- Setup a structure to keep track of which potions caused
		-- skill failure
		local AidPotTriesResetFunction = function()
							AidPotTriesLeft = { AidPotTriesPerLevel,
							AidPotTriesPerLevel, AidPotTriesPerLevel,
							AidPotTriesPerLevel }
		end
		AidPotTriesResetFunction()

		-- There is no current try
		AidPotAttempted = 0

		-- Setup a timeout to reset the tries
		if AidPotTriesReset > 0 then
			Timeout.add(AidPotTriesReset, AidPotTriesResetFunction, true)
		end

		-- Make sure we won't get any divide by 0 errors
		if AidPotIncrement == 0 then
			-- They asked for it! Its what every other 0
			-- amount turns into!
			AidPotIncrement = 100
		end

		-- Check if highest is lower than lowest
		if AidPotHighest < AidPotLowest then
			-- Swap 'em
			local temp = AidPotHighest
			AidPotHighest = AidPotLowest
			AidPotLowest = temp
		end

		-- Set the timeout for AidPot
		AidPotTimeout = 0

		-- Set the AidPotCheck function
		AidPotCheck = AI_AidPotion

		-- Add a function to our alchemist skills table that calls AidPotCheck
		AddAlchSkill(20,function(castAllowed)
			-- Call AidPotCheck
			return AidPotCheck(castAllowed)
		end)
	end

	-- Check if they want to use bolts automatically
	if AutoBoltSword then
		-- Set BoltLevelAttempted to 0, to prevent an error
		BoltLevelAttempted = 0

		-- Set the AutoBolt function (so custom scripts can hook it)
		AutoBoltCheck = AI_AutoBolt

		-- Add the auto-bolt function
		AddAlchSkill(30,function(castAllowed)
			-- Call AutoBoltCheck (to allow script to hook)
			return AutoBoltCheck(castAllowed)
		end)
	end

	-- Calculate the casting-time modifier
	CastModifier = 1.0 - (AlchemistDex/150)

	-- Check if they're wearing a phen
	if AlchemistPhen then
		-- Multiply the cast modifier by 125%
		CastModifier = CastModifier * 1.25

		-- Use the phen version of owner motion
		GetOwnerMotion = AI_GetOwnerMotion_Phen
	else
		-- Use the non-phen owner motion
		GetOwnerMotion = AI_GetOwnerMotion_NoPhen
	end

	-- Set the check alchemist skills function
	CheckAlchSkills = AI_CheckAlchSkills
end

-- Function to add skills to the alch skills priority table
function AddAlchSkill(priority,func)
	-- Check that func is a function
	if type(func) ~= "function" then
		return
	end

	-- Create the object we're adding
	local obj = { }
	obj.priority = priority
	obj.func = func

	-- Loop through our table finding a priority
	local i
	for i=Table.getn(AlchSkillsTable),1,-1 do
		-- Check the priority of the current function
		if priority >= AlchSkillsTable[i].priority then
			-- Insert our object
			Table.insert(AlchSkillsTable,i+1,obj)

			-- Don't loop anymore
			return
		end
	end

	-- Since we didn't find a place for it, add it at the front
	Table.insert(AlchSkillsTable,1,obj)
end

-------------------------------
-- Auto-Use Alchemist Skills --
-------------------------------

function AI_CheckAlchSkills(castAllowed)
	-- Check castAllowed
	if castAllowed == nil then
		castAllowed = true
	elseif not castAllowed then
		castAllowed = false
	end

	-- Loop through our table of alchemist skills
	--	** Idea for priority table from Katrina of RagnaInfo forums **
	local i
	for i=1,Table.getn(AlchSkillsTable) do
		-- Call the function and update the castAllowed
		castAllowed = AlchSkillsTable[i].func(castAllowed)
	end

	return castAllowed
end

----------------------------------
-- Auto-Homunculus Resurrection --
----------------------------------

-- (Idea by Spike Sydney)

-- Function to check (and use) Homunculus Resurrection if necessary
function AI_HomuResurrect(castAllowed)

	--[[
		Hidden option

		This specifies how long after it tries to Auto-Resurrect that
		it'll allow the Homunculus to move outside of the 2 tile range
	--]]
	local MaxDistanceDelay = 1000

	-- Get the ticks
	local ticks = GetTick()

	-- Check if its time to reset MaxDistance
	if MaxDistance ~= 0 and MaxDistanceTimeout < ticks then
		MaxDistance = 0
	end

	-- Check if our owner is sitting down.
	if OwnerMotion == MOTION_SIT then
		-- We can't do anything :(
		return castAllowed
	end

	-- Check if the homunculus is losing life
	if HomunHPDelta.iteration >= 0 and HomunHPDelta.second >= 0 then
		-- Not losing life, so why bother?
		return castAllowed
	end

	--[[ SPUsable automatically deducts 50sp from the Owner's SP pool
		so that other skills don't use it. If we do in fact have at
		least 50 SP, then we will break even at 0 or above
	--]]
	if SPUsable(OwnerID) < 0 then
		-- We don't have enough SP, allow aid pot
		return castAllowed
	end

	-- Check if the Homunculus's HP is below a certain amount
	if ActorHP(HomunID) >= AutoResHP then
		-- Check if we need to "intelligently" cast Homunculus Resurrection
		if AutoResSmart then
			-- Grab the amount of damage that the homunculus is taking per second/iteration
			local sec = HomunHPDelta.second
			local iter = HomunHPDelta.iteration
			local hp = GetV(V_HP,HomunID)

			-- Nullify any positives
			if iter > 0 then
				iter = 0
			end
			if sec > 0 then
				sec = 0
			end

			-- Make the absolute values
			sec = math.abs(sec)
			iter = math.abs(iter)

			-- Check the distance of the homunculus
			local dist = AADistance(HomunID,OwnerID) + 2

			-- Multiply the damage based on distance
			sec = sec * math.ceil(dist / 2)
			iter = iter * dist
			
			-- Check if the homunculus won't die within the time it'll get close enough
			-- to auto-resurrect itself
			if hp > iter and hp > sec then
				-- We probably won't die, see if we should aid pot
				return castAllowed
			end

			if hp < sec then
				Log(4,"Smart routine flagged auto-res (second delta)")
			else
				Log(4,"Smart routine flagged auto-res (iteration delta)")
			end
		else
			-- We aren't casting resurrection, so continue the allowed
			return castAllowed
		end

	else
		Log(4,"Dumb routine flagged auto-res")
	end

	-- Check if the Homunculus is in range of resurrection (range 2, we move to range 1)
	local dist = AADistance(HomunID,OwnerID)
	if dist > 1 then
		-- Check if we're not already moving towards our owner
		local seq = { SEQ_NONE }
		if List.size(AI_seq) > 0 then
			seq = AI_seq[AI_seq.first]
		end
		if seq[1] ~= SEQ_MOVE_TO_POS or seq[2] ~= OwnerID or seq[3] ~= 1 then
			-- Log it
			Log(1,"Moving within 1 tile of Owner, so we can Auto-Homunculus Resurrection")
	
			-- Create a sequence
			local seq2 = { SEQ_MOVE_TO_POS, OwnerID, 1, loglevel=6 }
	
			-- Begin the command
			local dummy
			dummy,seq2 = SEQ_AMove_Pos(seq2)
	
			-- Add it to the front of the sequence queue
			if seq2 ~= nil then
				List.pushleft(AI_seq,seq2)
			end
		end

		-- Check the distance of the homunculus
		local dist = AADistance(HomunID,OwnerID)

		-- Set MaxDistance (and it'll wear off after a delay)
		MaxDistance = 1
		MaxDistanceTimeout = ticks + math.floor(MaxDistanceDelay * dist / 3)

		-- While the homunculus is out of range of resurrection, allow aid pot
		if dist > 2 then -- disabled
			return castAllowed
		end
	end

	-- If the cast is not allowed, we're still moving closer for when it is allowed
	if not castAllowed then
		-- ... but we're going to pass on that casting isn't allowed

		-- Log it
		Log(4,"Not allowed to cast Homunculus Resurrection yet, still moving closer.")

		-- Return false
		return false
	end
	
	-- Make sure we're ready to cast
	if OwnerMotion ~= MOTION_CASTING and OwnerMotion ~= MOTION_SKILL then

		-- Log it
		Log(1,"Attempting to auto-cast Homunculus Resurrection")

		-- Use the skill
		SkillObject(HomunID, 5, 247, HomunID)

		-- Set our cast time and distance timeout
		OwnerSkillTimeout = ticks + (CastModifier * 3000) + AutoResDelay
		MaxDistanceTimeout = OwnerSkillTimeout + MaxDistanceDelay

	else
		-- Log it
		--Log(0,"Not casting Homunculus Resurrect because we're not ready yet")

		-- Set our distance timeout
		MaxDistanceTimeout = ticks + MaxDistanceDelay
	end

	-- Move the Homunculus to itself to prevent tile-lagging
	BypassMove(HomunID,GetV(V_POSITION,HomunID))

	-- Make sure they stay within 1 tile
	MaxDistance = 1

	-- We're using Resurrection, so don't try to Aid Pot
	return false
end


---------------------
-- Auto-Aid-Potion --
---------------------


-- Function to check (and use) Aid Potion if necessary
function AI_AidPotion(castAllowed)
	--[[ NOTE: If castAllowed is false, it will only check to see
		if a previous Aid Potion was successful
	--]]

	-- Get the tick count
	local ticks = GetTick()

	-- Check if we've had a success yet
	local success = false
	if AidPotAttempted > 0 then
		-- Check if our HP delta is positive (or owner's motion is PP)
		if HomunHPDelta.iteration > 0 or OwnerMotion == MOTION_PP then
			-- It was, so we probably were successful
			success = true
		else
			-- Our HP delta wasn't positive, so check if
			-- it was better than the last few
			
			-- Loop through the history of HP deltas
			local i,count,hpD = 0,0,HomunHPDelta.iteration
			for i=AidPotHistory.last-1,AidPotHistory.first,-1 do
				-- Get the history entry
				local t = AidPotHistory[i]
				
				-- Check if the history was aid-potted
				if not t[2] then
					-- It wasn't, check if we beat it
					if t[1] >= hpD then
						-- It wasn't better, so break
						break
					end
					
					-- Otherwise, add one to our count
					count = count + 1
				end
			end
			
			-- See if we beat the last 3 or more (that weren't aided)
			if count >= 3 then
				-- Count it as a success
				success = true
			end
		end

		-- Check if we found success
		if success then
			-- Log it
			Log(0,"Aid Potion was apparently successful.")
				
			-- Reset the tries counter (we obviously haven't run out)
			AidPotTriesLeft[AidPotAttempted] = AidPotTriesPerLevel

			-- We'll count it as a success
			AidPotAttempted = 0
			
			-- Reset our timeout as well
			AidPotTimeout = 0
		end
	end
	
	-- Add the HP delta to our history list
	List.pushright(AidPotHistory,{ HomunHPDelta.iteration, success })
	
	-- Trim it down if its too long
	while List.size(AidPotHistory) > 40 do
		List.popleft(AidPotHistory)
	end

	-- Check if our owner is still waiting for cast delay
	if ticks < AidPotTimeout and AidPotTimeout ~= 0 then
		-- Still waiting
		return castAllowed
	end

	-- Check if we failed (attempted was never reset to 0)
	if AidPotAttempted > 0 then
		-- Subtract a try from that level
		AidPotTriesLeft[AidPotAttempted] = AidPotTriesLeft[AidPotAttempted] - 1

		-- Log it
		Log(0,"Aid Potion appears to have been unsuccessful, %d tries left for Aid Potion level %d",
			AidPotTriesLeft[AidPotAttempted],AidPotAttempted)
		
		-- Set attempted to 0
		AidPotAttempted = 0
	end

	-- If we're not allowed to cast...
	if not castAllowed then
		-- ... Then we'll pass it on
		return false
	end

	-- Check if our owner is sitting down.
	if OwnerMotion == MOTION_SIT then
		-- We can't do anything :(
		return true
	end

	-- Check if our owner walked recently
	if OwnerLast.Move + AidPotWaitAfterMove >= ticks then
		-- Not time to use aid pot yet
		return true
	end

	-- Check our HP
	if ActorHP(HomunID) >= AidPotHP then
		-- Not time to use aid pot yet
		return true
	end

	-- Get the amount of SP that we have
	if SPUsable(OwnerID) < 1 then
		-- We don't have enough SP
		return true
	end

	-- Check if they're in range to get aid-potted
	if AADistance(HomunID,OwnerID) > 10 then
		-- Check if they want us to try to move closer
		if AidPotMoveCloser then
			-- Get the next sequence
			local seq
			if List.size(AI_seq) > 0 then
				seq = AI_seq[AI_seq.first]
			else
				seq = { SEQ_NONE }
			end

			-- Check if its a move to our owner
			if (seq[1] == SEQ_MOVE_TO or seq[1] == SEQ_MOVE_TO_POS or
				seq[1] == SEQ_MOVE_TO_NEG) and seq[2] == OwnerID and
				math.abs(seq[3]) <= 6
			then
				-- Its a move to our owner, just wait for it
				return true
			end

			-- Add a movement to our owner
			List.pushleft(AI_seq,{ SEQ_MOVE_TO, OwnerID, 5 } )
		end

		-- They're out of range
		return true
	end

	-- Seems like its time to use it, check the number
	-- of attackers of us

	local enemies = ActorEnemies(HomunID)
	local numEnemies = Table.getn(enemies)

	-- Determine the level of aid potion to use
	local level = math.floor(numEnemies / AidPotIncrement)

	-- Check if its time to use it
	if level == 0 then
		-- Not time to use aid pot yet
		return true
	end
	
	-- Add our lowest level (so it really is increment)
	level = level + AidPotLowest - 1

	-- Make sure its not higher than our highest
	if level > AidPotHighest then
		-- Set it to highest
		level = AidPotHighest
	end

	-- Make sure we have the potions required
	local start,highest = level,AidPotHighest
	while level <= highest do
		-- Check if we have potions of this level
		if AidPotTriesLeft[level] > 0 then
			-- We do, so try to use 'em
			break
		end
		
		-- Check if incrementing would put us past our highest
		if level + 1 > AidPotHighest then
			-- Check if we can start from the bottom
			if start ~= AidPotLowest then
				-- Set our new highest
				highest = start - 1
				
				-- Set our new level
				level = AidPotLowest
				
				-- Set our new start
				start = level
			else
				-- We can't, just exit the loop
				break
			end
		else
			-- Otherwise, increment the potion level by 1
			level = level + 1
		end
	end
	
	-- Check if we have any potions
	if AidPotTriesLeft[level] > 0 then
		-- Check if we're already casting something
		if OwnerMotion ~= MOTION_CASTING and OwnerMotion ~= MOTION_SKILL then
			-- Log it
			Log(0,"Attempting to automatically use Aid Potion, level %d.",
				level)

			-- Use the skill
			SkillObject(HomunID, level, 231, HomunID)

			-- Set the timeouts (cast delay)
			--OwnerSkillTimeout = ticks+AidPotWaitTime
			AidPotTimeout = OwnerSkillTimeout

			-- Reset the success variable
			AidPotAttempted = level
		end

		-- Don't allow auto-bolting while we're aid potting
		return false
	end

	-- Since we're not aid-potting, allow bolts to fire
	return true
end

---------------
-- Auto-Bolt --
---------------

-- Function to use a bolt-sword (Ice Falchion/Fireblend)
--		(idea from a variety of people)
function AI_AutoBolt(castAllowed)
	-- Check if we're sitting down
	if OwnerMotion == MOTION_SIT then
		-- Can't use any skills :(
		return true
	end

	-- Get the current ticks
	local ticks = GetTick()

	-- Check if we previously tried casting a bolt
	if BoltLevelAttempted ~= 0 and MaxDistance == 0 then

		-- Check if we're casting
		if OwnerMotion == MOTION_CASTING then
			-- We're casting now, set our timeout
			
			-- Formula modified from Fallen~Angel
			OwnerSkillTimeout = ticks +
				-- Cast Time
				(CastModifier * 700 * BoltLevelAttempted) +

				-- Cast delay
				800 + (200 * BoltLevelAttempted)

			-- Log it
			Log(2,"Using a sword-bolt (level %d) against Actor #%d.",
				BoltLevelAttempted,BoltTargetted)

			-- Set the time which we last casted a sword's bolt
			LastBoltSword = ticks

			-- Increment the number of times we've casted at this monster
			AlchTarget.Skill = AlchTarget.Skill + 1

			-- Reset bolt level
			BoltLevelAttempted = 0
			BoltTimeout = 0

			-- Don't allow casting
			return false

		elseif ticks > BoltTimeout then

			-- Bolt failed
			BoltLevelAttempted = 0
			BoltTimeout = 0

		end

	end

	-- Check to see if we're allowed to cast
	if not castAllowed then
		return false
	end

	-- Check if we're still in a delay time for auto-res
	if MaxDistance ~= 0 then
		return true
	end

	-- Check if we have a target for our alchemist
	if AlchTarget.ID == -1 then
		return true
	end

	-- Try not to break autoattack (ctrl+click, or /nc and click)
	if OwnerLast.Attack + AutoBoltDelay >= ticks then
		-- Too early to cast, they attacked recently
		return true
	end

	-- Check if we've moved recently
	if OwnerLast.Move + AutoBoltDelay >= ticks then
		-- Too early to cast
		return true
	end

	-- Check if we're already casting
	if	OwnerMotion == MOTION_SKILL or
		OwnerMotion == MOTION_CASTING or
		OwnerMotion == MOTION_PP or
		OwnerMotion == MOTION_SLIMPP then

		-- We don't want to cast right now
		return true

	end

	-- Check if we've casted too many times against this one
	if MSwBoltMaxCasts(Target.ID) <= AlchTarget.Skill then
		-- Don't cast
		return true
	end

	-- Check if its too early to cast against this monster
	if LastBoltSword + MSwBoltTimeout(AlchTarget.ID) > ticks then
		-- Too early to cast against it
		return true
	end

	-- Get the amount of SP we're allowed to use
	local sp = SPUsable(OwnerID)

	-- Get the max level we can use against our target
	local max = MSwBoltMaxLevel(AlchTarget.ID)

	-- Make sure the maximum level is a sane value
	if max < 1 then
		-- Don't use the bolt
		return true
	elseif max > 3 then
		-- Set the max at level 3
		max = 3
	end

	-- Get the minimum level we should cast of our sword bolts
	local min = AutoBoltMinLevel

	-- Make sure max isn't below min
	if max < min then
		min = max
	end

	-- Determine which level to use
	local level = 0
	local i
	for i=max,min,-1 do
		if sp >= 10 + (2 * i) then
			level = i
			break
		end
	end

	-- Check if we are going to use a valid level
	if level == 0 then
		-- Don't cast a bolt
		return true
	end

	-- Attempt to use Fire Bolt / Ice Bolt
	SkillObject(HomunID, level, 14, AlchTarget.ID)
	SkillObject(HomunID, level, 19, AlchTarget.ID)

	-- Set the level we're attempting
	BoltLevelAttempted = level
	BoltTimeout = ticks + AutoBoltDelay
	BoltTargetted = AlchTarget.ID

	-- Don't allow any other skills to be used
	return false
end
