--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

----------------------
-- Attack Functions --
----------------------

-- Simple attack function
function AI_Attack()
	Attack(HomunID,Target.ID)
end

-- Attack (with skills too) function
function AI_SkilledAttack()
	-- Send a skill attack
	SendAttackSkill()

	-- After everything else, try to do a normal attack
	Attack(HomunID,Target.ID)
end

-- Send an attack skill command
function SendAttackSkill()
	-- Check if skills are allowed
	if not MSkillsAllowed(Target.ID) then
		-- If they aren't allowed, return false
		return false
	end

	-- Check if we're ready to cast a skill
	if not AttackSkillReady() then
		-- Not time to cast a skill yet
		return false
	end

	-- Get a skill to use
	local skill,level = CheckAttackSkill()

	-- Check if we got a skill
	if skill ~= 0 then
		-- Use the skill
		UseAutoSkill(level,skill)
		
		-- We used a skill, so return true
		return true
	end

	-- If we didn't use a skill, return false
	return false
end

-- Function to check if we should use a skill
function AI_CheckAttackSkill()
	-- Get the range
	local range = AADistance(HomunID,Target.ID)

	-- Check the type
	local type
	if range <= HomunRange then
		type = SKILL_TYPE_MELEE
	else
		type = SKILL_TYPE_RANGED
	end

	--[[
		If "type" is melee, it'll check melee and ranged skills,
		otherwise the only other type that should be used is
		ranged, and it will only check those
	--]]

	-- First things first, check if its allowed
	if not SkillAllowed(type) then
		-- No skill, sorry
		return 0,0
	end

	-- Check if we're allowed to use a skill against that actor yet
	if not AttackSkillReady() then
		-- Not yet
		return 0,0
	end

	-- Grab the amount of SP that we're allowed to use
	local sp = SPUsable(HomunID)

	-- Grab the max skill level that we should use
	local maxLevel = MSkillMaxLevel(Target.ID)

	-- Decide the min skill level that we should use
	local minLevel = MinAttackSkillLevel

	-- Verify that min isn't greater than max
	if minLevel > maxLevel then
		minLevel = maxLevel
	end

	-- Check if we should do melee skills as well
	local i,skill,j
	if type == SKILL_TYPE_MELEE and Table.getn(AttackSkills) > 0 then
		-- Loop through all our melee skills
		for i,skill in ipairs(AttackSkills) do
			-- Loop through the SP table
			for j=maxLevel,minLevel,-1 do
				-- If we have enough SP for it, use it
				if sp >= skill.sp[j] then
					-- Use the skill
					return skill.id,j
				end
			end
		end
	end

	-- Check if we have any ranged skills to try and use
	if Table.getn(RangedSkills) < 1 then
		-- We don't, so no skill this time
		return 0,0
	end

	-- Check if range matters
	if range == nil then
		-- Doesn't matter, set it super high
		range = 100
	end

	-- Now check for a ranged skill to use
	for i,skill in ipairs(RangedSkills) do
		-- Check if the skill is in range
		if range <= skill.range then
			-- Loop through the SP table
			for j=maxLevel,minLevel,-1 do
				-- Check its SP
				if sp >= skill.sp[j] then
					-- We got our skill, return it
					return skill.id,j
				end
			end
		end
	end

	-- No valid skills, return 0,0
	return 0,0
end

-- Function to check if we should use a skill (at any range)
function AI_CheckAttackSkill_NoRange()
	-- Check the range from the owner
	if AADistance(OwnerID,Target.ID) > IgnoredRangeFromOwner then
		-- Not in range of the alchemist
		return 0,0
	end

	-- Check if its allowed
	if AADistance(HomunID,Target.ID) <= HomunRange + 1 then
		if not SkillAllowed(SKILL_TYPE_MELEE) then
			return 0,0
		end
	elseif not SkillAllowed(SKILL_TYPE_RANGED) then
			return 0,0
	end

	-- Check if we're allowed to use a skill against that actor yet
	if not AttackSkillReady() then
		-- Not yet
		return 0,0
	end

	-- Grab the amount of SP that we're allowed to use
	local sp = SPUsable(HomunID)

	-- Grab the max skill level that we should use
	local maxLevel = MSkillMaxLevel(Target.ID)

	-- Decide the min skill level that we should use
	local minLevel = MinAttackSkillLevel

	-- Verify that min isn't greater than max
	if minLevel > maxLevel then
		minLevel = maxLevel
	end

	-- Now check for a ranged skill to use
	for i,skill in ipairs(RangedSkills) do
		-- Loop through the SP table
		for j=maxLevel,minLevel,-1 do
			-- Check its SP
			if sp >= skill.sp[j] then
				-- We got our skill, return it
				return skill.id,j
			end
		end
	end

	-- No valid skills, return 0,0
	return 0,0
end

-- Function to check if its time to use a skill again
function AI_AttackSkillReady()
	-- Check if we're allowed to use skills against this monster
	if not MSkillsAllowed(Target.ID) then
		-- Not ready
		return false
	end

	-- Check if we're allowed another cast
	if MSkillMaxCasts(Target.ID) <= Target.Skill then
		-- Not allowed to cast any more skills against this guy
		return false
	end

	-- Check if theres still a delay
	if LastAttackSkill ~= 0 and
		GetTick() < (MSkillTimeout(Target.ID) + LastAttackSkill) then

		-- We're not ready yet
		return false
	end

	-- If it wasn't determined that we're not ready, then we are ready
	return true
end

-- Function to use a skill
function AI_UseSkill(level,type)
	if not AttackSkillReady() then
		-- Not time to use a skill
		return false
	end

	-- Log it
	Log(2,"Automatically using skill #%d (level %d) against actor #%d",
		type,level,Target.ID)

	-- Use the skill
	SkillObject(HomunID,level,type,Target.ID)

	-- Increment the number of times we've casted at them
	Target.Skill = Target.Skill + 1

	-- Reset the timeout (and add the cast delay)
	LastAttackSkill = GetTick() + AttackSkillDelay(type,level)
end

-----------------------------
-- Skill Support Functions --
-----------------------------

-- Add an attack skill to our lists
function AddAttackSkill(id,sptable,rangesuggest,delayfunc)
	-- Check if we're ignoring attack skill ranges
	local range
	if IgnoreAttackSkillRange then
		-- Set the range to 40 tiles
		range = 40
	else
		-- Check if we have a range suggestion
		if rangesuggest == nil then
			-- Get the range of the skill
			range = GetV(V_SKILLATTACKRANGE,HomunID,id)
		else
			-- Use the suggestion
			range = rangesuggest
		end
	end
	
	-- Check if the range is our lowest one yet
	if range < HomunSkillRange then
		HomunSkillRange = range
	end

	-- Make sure the sptable is valid
	local i
	for i=5,1,-1 do
		-- If there is an empty space in the sp usage table...
		if sptable[i] == nil then
			-- Fill it with a large number
			sptable[i] = 100 * 100
		end
	end

	-- Create the object
	local obj = { }
	obj.id = id
	obj.sp = sptable
	obj.range = range
	obj.delayFunc = delayfunc

	-- Check the delay function
	if type(obj.delayFunc) ~= "function" then
		-- Set it to zero return
		obj.delayFunc = ZeroReturn
	end

	-- Check the range
	if range <= (HomunRange+1) then
		-- Its a melee skill
		Table.insert(AttackSkills,obj)
	else
		-- Its a ranged skill
		Table.insert(RangedSkills,obj)
	end

	-- Add it to all homunculus skills
	AttackSkillsByID[id] = obj
end

-- Function to determine the cast delay a homunculus will suffer from casting
--	an attack skill
function AttackSkillDelay(skill,level)
	-- Check if we have a skill object for this skill
	if AttackSkillsByID[skill] ~= nil then
		-- Use its delay function
		return AttackSkillsByID[skill].delayFunc(level)
	end

	-- Don't have it, return 0
	return 0
end

-- Determine if a skill is allowed or not
function AI_SkillAllowed(type)
	-- If our owner is sitting, we can't cast //anything//
	if GetV(V_MOTION,OwnerID) == MOTION_SIT then
		-- Nope, no can do :(
		return false
	end

	-- Check if we have enough SP
	if ActorSP(HomunID) < MinSP then
		-- Will use too much SP, so its not allowed
		return false
	end

	-- Check if we're passive
	if Mode == MODE_PASSIVE then
		-- Check if we're trying to use an attack skill
		if type == SKILL_TYPE_MELEE or type == SKILL_TYPE_RANGED then
			-- Only allow it if the user okays it
			if not PassiveSkills then
				-- User didn't want it happening
				return false
			end
		end
	end

	-- Check if we're pausing attack skills for now
	if PauseAttackSkill and (type == SKILL_TYPE_MELEE or 
		type == SKILL_TYPE_RANGED) then

		-- Not allowed
		return false
	end

	-- Check the type of skill
	if type == SKILL_TYPE_MELEE and MeleeSkills then
		-- Its allowed
		return true
	elseif type == SKILL_TYPE_RANGED and ChaseSkills then
		-- Its allowed
		return true
	elseif type == SKILL_TYPE_SUPPORT and SupportSkills then
		-- Its allowed
		return true
	elseif type == SKILL_TYPE_HEAL and HealSkills then
		-- Its allowed
		return true
	end

	-- Otherwise its not allowed
	return false
end

--------------------
-- Target Setting --
--------------------

-- Function to set the target
function SetTarget(id)
	-- Set the previous
	Target.Previous = Target.ID

	-- Make sure its a number
	if type(id) ~= "number" then
		-- Log it
		Log(0,"Invalid ID passed to SetTarget(%s).",tostring(id))

		-- Set it to -1
		id = -1
	end

	-- Check if its the same monster
	if id == Target.ID then
		return
	end

	-- Set the new target
	Target.ID = id

	-- Set the target lock
	Target.Lock = false

	-- Set the attack skill counter
	Target.Skill = 0

	-- Reset the attack log
	ResetAttackLog()

	-- Set the start tick
	Target.StartTick = GetTick()
end

-- Function to set the alchemist target
function AlchSetTarget(id)
	-- Set the previous
	AlchTarget.Previous = AlchTarget.ID

	-- Make sure its a number
	if type(id) ~= "number" then
		-- Log it
		Log(0,"Invalid ID passed to AlchSetTarget(%s).",tostring(id))

		-- Set it to -1
		id = -1
	end

	-- Check if its the same monster
	if id == AlchTarget.ID then
		return
	end

	-- Set the new target
	AlchTarget.ID = id

	-- Set the target lock
	AlchTarget.Lock = false

	-- Set the attack skill counter
	AlchTarget.Skill = 0

	-- Set the start tick
	AlchTarget.StartTick = GetTick()
end

-----------------------------------
-- Homunculus Target Acquisition --
-----------------------------------

-- Function to acquire a target for the homunculus
function HomuAcquireTarget()
	-- Check if the target is locked, but we have no attack command
	if Target.Lock and not InSequence(SEQ_ATTACK) and
		not InSequence(SEQ_KITE) then
		-- Take target lock off
		Target.Lock = false
	end

	-- Check if we have a target-lock (and we don't ignore it)
	if Target.Lock and Target.ID ~= -1 and not AcquireWhileLocked then
		-- Reset the swap counter
		Target.Swap = 0

		-- Don't try to acquire another target
		return
	end

	-- Default our action going to the back
	local addAtBack = true

	-- Check if the next action is going to be the wrong type
	if List.size(AI_seq) > 0 and Target.ID ~= -1 and Target.Lock then
		-- Check if the next action is going to be the wrong type
		if (AI_seq[AI_seq.first][1] == SEQ_ATTACK and MKite(Target.ID)) or
			(AI_seq[AI_seq.first][1] == SEQ_KITE and not MKite(Target.ID)) then
			-- It will be the wrong type

			-- Don't add to the back
			addAtBack = false

			-- Pop the wrong type off the front
			List.popleft(AI_seq)
		end
	end
	
	-- Determine which function to use to acquire targets
	local AcquireTargetFunc = PassiveAcquire
	if Mode == MODE_AGGRO then
		AcquireTargetFunc = AggroAcquire
	end

	-- Determine if there is a function to override homunculus target acquisition
	if type(HomunAcquireOverride) == "function" then
		-- Use it
		AcquireTargetFunc = HomunAcquireOverride
	end

	-- Save our previous target
	local saved = Target.Previous

	-- Acquire a target, and check it
	if AcquireTargetFunc(false,false,false) == -1 then
		-- No target, so return right away
		return
	end

	-- Hold position is useless now
	HoldPosition = false

	-- Ensure we have an attack command on the AI_seq
	if (not InSequence(SEQ_ATTACK) and not MKite(Target.ID)) or
		(not InSequence(SEQ_KITE) and MKite(Target.ID)) then

		-- Check the monster's setting
		local seq = { SEQ_ATTACK }
		if MKite(Target.ID) then
			seq = { SEQ_KITE }
		end

		-- Check if we're supposed to add to the back
		if addAtBack then
			List.pushright(AI_seq,seq)
		else
			List.pushleft(AI_seq,seq)
		end
	end

	-- Check if we have a different target
	if Target.ID ~= Target.Previous then
		-- Log our new target
		Log(1,"Now targetting actor #%d (previously actor #%d)",
			Target.ID,Target.Previous)
	end

	-- Check if our target is the previous one
	if saved == Target.ID and
		saved ~= Target.Previous then
		-- Seems like we swapped targets

		-- Increment our swap counter
		Target.Swap = Target.Swap + 1

		-- Check if we've swapped too often
		if Target.Swap == 3 then
			-- Log it
			Log(2,"Locking on to actor #%d, due to too many target swaps",
				Target.ID)
			
			-- Set target lock
			Target.Lock = true
		end
	else
		-- Reset our swap counter
		Target.Swap = 0
	end
end

-- Determine if we should assist our owner
function AI_AssistOwnerAcquire(reused,foralch,isolated)
	-- Get our owner's target
	local targ = -1
	if isolated ~= true then
		-- No function to run before this one
	end

	-- Check if a previous function failed to find a target
	if targ == -1 then
		-- Get the target of our owner
		targ = GetV(V_TARGET,OwnerID)

		-- Check if there is a monster
		--if targ == -1 then
			-- No monster, return nothing
		--end

		-- Check if the monster is too far away, or dead
		if targ ~= -1 and
			(GetV(V_MOTION,targ) == MOTION_DEAD or
			not ActorIsInSight(targ)) then

			-- Return no monster
			targ=-1
		end
	
		-- Check if our owner attacked recently
		if targ ~= -1 and OwnerLast.Attack + 500 < GetTick() then
			-- They didn't attack recently, don't assist
			targ=-1
		end

		-- Check if the monster is allowed
		if targ ~= -1 and
			(IsMonster(targ) ~= 1 or
			not MonsterAllowed(actor) or
			IsIgnored(actor)) then
			-- Monster isn't allowed, return -1
			targ=-1
		end
	end

	-- Check if we should update the Target structure
	if reused ~= true then
		-- Check if its for the alchemist
		if foralch ~= true then
			-- Its for homunculus
			SetTarget(targ)
		else
			-- Its for alchemist
			AlchSetTarget(targ)
		end
	end
	
	return targ
end

-- Determine which enemies are attacking us
function AI_PassiveAcquire(reused,foralch,isolated)
	-- Check to see if we should assist our owner
	local target = -1
	if isolated ~= true then
		target = AssistAcquire(true,foralch)
	end

	-- Check if we have a target that our owner is attacking
	if target == -1 then
		-- Don't assist
		
		-- Identify enemies attacking our owner
		local ownerAttackers = IgnoredSieve(ActorEnemies(OwnerID))

		-- Identify enemies attacking the homunculus
		local homunAttackers = IgnoredSieve(ActorEnemies(HomunID))
	
		-- Check if we care about our friends
		local friendAttackers = Table.new()
		local friendSpecific = Table.new()
		if DefendFriends then
			-- Variables to keep count of the most number of enemies
			--	targetting each friend
			local maxAtkers,maxAtkersActor = -1,-1

			-- Loop through each friend who is on the screen
			local i,friendactor
			for friendactor,i in pairs(PermFriendsOnScr) do
				-- Call upon a support function
				friendAttackers,maxAtkers,maxAtkersActor,friendSpecific =
					PassiveAcquireFriendSupport(friendAttackers,maxAtkers,
						maxAtkersActor,friendSpecific,friendactor)
			end

			-- Loop through each temporary friend
			for friendactor,i in pairs(TempFriends) do
				-- Call upon a support function
				friendAttackers,maxAtkers,maxAtkersActor,friendSpecific =
					PassiveAcquireFriendSupport(friendAttackers,maxAtkers,
						maxAtkersActor,friendSpecific,friendactor)
			end
		end
	
		-- Default to prioritizing monsters on homunculus
		local defendTable,defendPrio,defendAbove = homunAttackers,PrioritizeHomunDefense,false

		-- Check if the number of monsters attacking our homunculus is above
		-- 	the prioritize number
		if Table.getn(defendTable) >= defendPrio then
			-- It is
			defendAbove = true
		end

		-- Check if we should prioritize our owner's defense
		if Table.getn(ownerAttackers) >= PrioritizeOwnerDefense then
			-- Owner defense is important

			-- Default to defending our owner
			local defendOwner = true

			-- Check to see if the previous table has met its
			-- 	prioritization quota
			if defendAbove then
				-- Check to see if the other table was more important
				if defendPrio > PrioritizeOwnerDefense then
					-- Owner isn't more important than the current
					defendOwner = false

				-- Check to see if its equal importance
				elseif defendPrio == PrioritizeOwnerDefense and
					-- Check which actor has more attackers
					Table.getn(ownerAttackers) <= Table.getn(defendTable) then

					-- The previous defense table has more attackers,
					-- so we'll stick with that one
					defendOwner = false
				end
			end

			-- Check if we still want to defend our owner
			if defendOwner then
				-- Set the table
				defendTable = ownerAttackers

				-- Set the prioritization
				defendPrio = PrioritizeOwnerDefense

				-- We were, in fact, above our prioritization
				defendAbove = true
			end

		else
			-- Owner defense is not important, check if the previous one was

			if not defendAbove then
				-- It wasn't important, so add our monsters to the
				-- monsters in the list
				Table.insert(defendTable,ownerAttackers)

				-- Since defendAbove is false, defendPrio doesn't matter
			end
		end

		-- Check if we're supposed to defend our friends
		if DefendFriends then
			-- Check if we should prioritize a specific friend's defense over owner/homun's
			if Table.getn(friendSpecific) >= PrioritizeFriendDefense then

				-- Default to defending the friend
				local defendFriend = true

				-- Check if they were already above a prioritization
				if defendAbove then
					-- Check to see if the other table was more important
					if defendPrio > PrioritizeFriendDefense then
						-- Friend isn't more important than the current
						defendFriend = false
	
					-- Check to see if its equal importance
					elseif defendPrio == PrioritizeFriendDefense and
						-- Check which actor has more attackers
						Table.getn(friendSpecific) <= Table.getn(defendTable) then
	
						-- The previous defense table has more attackers,
						-- so we'll stick with that one
						defendFriend = false
					end
				end

				-- Check if we still want to defend our friend
				if defendFriend then
					-- Set the table
					defendTable = friendSpecific

					-- Set the prioritization
					defendPrio = PrioritizeFriendDefense

					-- We're above our prioritization
					defendAbove = true
				end

			end

			-- Check if the table ever met a prioritization
			if not defendAbove then
				-- It never met the prioritization, so add all of
				-- the monsters attacking any of our friends to the list
				Table.insert(defendTable,friendAttackers)
			end
		end

		-- Determine if its for the alchemist
		local actor = HomunID
		if foralch == true then
			actor = OwnerID
		end

		-- Get the closest monster of the highest priority
		--	and do the kited monsters last
		target = ATNearest(actor,MPrioritySieve(MKiteSieve(defendTable)))
	end

	-- Check if we should update the Target structure
	if reused ~= true then
		-- Check if this is for our alchemist
		if foralch ~= true then
			-- Its for homunculus
			SetTarget(target)
		else
			-- Its for the alchemist
			AlchSetTarget(target)
		end
	end

	return target
end

-- Support function for the above PassiveAcquire script
function PassiveAcquireFriendSupport(friendAttackers,maxAtkers,maxAtkersActor,maxAtkersList,friendactor)
	--[[

		 Disabled => This will unnecessarily take exponentially more time

	-- Identify enemies attacking them
	friendAttackers = ActorEnemies(friendactor,friendAttackers)

	-- Remove ignored monsters
	friendAttackers = IgnoredSieve(friendAttackers)

	--]]

	-- Make sure we're sane
	if type(friendactor) ~= "number" and IsFriend(friendactor) then
		-- We're not
		return friendAttackers,maxAtkers,maxAtkersActor,maxAtkersList
	end

	-- Get the enemies attacking this specific actor
	local frSpecific = IgnoredSieve(ActorEnemies(friendactor))

	-- Get the delta of attackers
	local atkers = Table.getn(frSpecific)

	-- Merge the friend specific into the total friend attackers
	Table.insert(friendAttackers,frSpecific)

	-- See if its more than our current max
	if atkers > maxAtkers then
		-- Set the new max
		maxAtkers = atkers
		maxAtkersActor = friendactor
		maxAtkersList = frSpecific

	-- Check if its equal to our current max
	elseif atkers == maxAtkers then

		-- Dist1
		local dist1 = AADistance(HomunID,friendactor)

		-- Dist2
		local dist2 = AADistance(HomunID,maxAtkersActor)

		-- Check if our current actor is closer
		if dist1 < dist2 then

			-- Set the new max
			maxAtkers = atkers
			maxAtkersActor = friendactor
			maxAtkersList = frSpecific
		end
	end

	-- Pass back the data to the passive acquire function
	return friendAttackers,maxAtkers,maxAtkersActor,maxAtkersList
end

-- Determine the closest target to attack
function AI_AggroAcquire(reused,foralch,isolated,extraCriteria)
	-- First and foremost, defend ourselves (unless this function
	--	is isolated)
	local target = -1
	if isolated ~= true then
		target = PassiveAcquire(true,foralch)
	end

	-- Check if we got a target thats attacking us
	if target == -1 then
		-- Loop through all actors, determining which are enemies
		local enemies = Table.new()
		local i,actor,motion
		for actor,i in pairs(ActorTargets) do
			-- Make sure the actor isn't homunculus or owner
			if actor ~= OwnerID and actor ~= HomunID then
				-- Get the motion of the actor
				motion = GetV(V_MOTION,actor)

				-- Make sure the actor is on the screen and not dead
				if motion ~= MOTION_DEAD and ActorIsInSight(actor) then
					-- Check if its a valid monster
					if IsMonster(actor) == 1 and
						MonsterAllowed(actor) and
						not IsFriend(actor) and
						not IsIgnored(actor) and
						not MNoAggroAgainst(actor) and
							(type(extraCriteria) ~= "function" or
							extraCriteria(actor)) and
						not KillSteal(actor)
					then
						-- Add the monster to the table
						Table.insert(enemies,actor)
					end
				end
			end
		end

		-- Determine if its from the alchemist or not
		actor = HomunID
		if foralch == true then
			actor = OwnerID
		end

		-- Find the nearest enemy, from the list of highest priority
		target = ATNearest(actor,MPrioritySieve(enemies))
	end

	-- Check if we should update the Target structure
	if reused ~= true then
		-- Check if this is for our alchemist
		if foralch ~= true then
			-- Its not
			SetTarget(target)
		else
			-- Its for alchemist's target
			AlchSetTarget(target)
		end
	end

	return target
end

-- Function to reuse the aggro acquire, but only to grab kited monsters
--	(also an example for HomunAcquireOverride)
function AI_AggroKitesOnly(reused,foralch,isolated)
	-- Check if we're in passive mode
	if Mode == MODE_PASSIVE then
		-- Reuse the aggro code above, but make it aggro only kited
		--	monsters when its not defending
		return AI_AggroAcquire(reused,foralch,isolated,MKite)
	else
		-- Aggro any monster while not defending
		return AI_AggroAcquire(reused,foralch,isolated)
	end
end

----------------------------------
-- Alchemist Target Acquisition --
----------------------------------

function AlchAcquireTarget()
	-- See if we're locked onto our target
	if AlchTarget.Lock and not AcquireWhileLocked then
		-- Check to make sure the target is alive and on screen
		if not ActorIsInSight(AlchTarget.ID) or
			GetV(V_MOTION,AlchTarget.ID) then

			-- Unset target lock
			AlchTarget.Lock = false
		else
			-- Its on screen, so don't acquire
			return
		end
	end

	-- Determine which function to use to acquire targets
	local AcquireTargetFunc = AlchAcquire

	-- Determine if there is a function to override alchemist target acquisition
	if type(AlchAcquireOverride) == "function" then
		-- Use it
		AcquireTargetFunc = AlchAcquireOverride
	end

	-- Save our previous target
	local saved = AlchTarget.Previous

	-- Acquire a target, and check it
	if AcquireTargetFunc(false,true,false) == -1 then
		-- No target, so return right away
		return
	end

	-- Check if we have a different target
	if AlchTarget.ID ~= AlchTarget.Previous then
		-- Log our new target
		Log(1,"Now targetting actor #%d (previously actor #%d) for alchemist abilities",
			AlchTarget.ID,AlchTarget.Previous)
	end

	-- Check if our target is the previous one
	if saved == AlchTarget.ID and
		saved ~= AlchTarget.Previous then
		-- Seems like we swapped targets

		-- Increment our swap counter
		AlchTarget.Swap = AlchTarget.Swap + 1

		-- Check if we've swapped too often
		if AlchTarget.Swap == 3 then
			-- Log it
			Log(2,"Alchemist locking on to actor #%d, due to too many target swaps",
				AlchTarget.ID)

			-- Set target lock
			AlchTarget.Lock = true
		end
	else
		-- Reset our swap counter
		AlchTarget.Swap = 0
	end
end

-- Function to use the same target as the homunculus
function AI_AlchSupportAcquire(reused,foralch,isolated)
	-- Make sure its for the alchemist
	if foralch ~= true then
		return -1
	end

	-- Get the homunculus's target
	local target = Target.ID

	-- Check to make sure the target is attacking us
	--	(so we don't aggro to the alchemist)
	local targsTarg = GetV(V_TARGET,target)
	if targsTarg ~= OwnerID and targsTarg ~= HomunID and
		not IsFriend(targsTarg) then

		-- Don't attack it yet
		target = -1
	end

	-- Check if we should update the Target structure
	if reused ~= true then
		AlchSetTarget(target)
	end

	return target
end

----------------------------
-- Attack Log Maintaining --
----------------------------

-- Trim down the attack log against the current target
function AI_TrimAttackLog()
--[[   Homunculus's GetV(V_MOTION) never shows MOTION_ATTACK. Ever.

	HOWEVER! If we check to see if our target has been damaged, then we
	may use that as a pseudo-MOTION_ATTACK for our homunculus

--]]

	local atkTime = 0

	-- Get the attack time
	--atkTime = HomunLast.Attack

	-- Fudge the attack time
	atkTime = ActorDamageTimes[Target.ID]

	-- Check if they've ever been damaged
	if atkTime == 0 then
		-- Get the time we started combat
		atkTime = Target.StartTick
	end

	-- Trim the attack log
	while List.size(Target.AttackLog) > 0 do
		-- Check the first entry
		local entry = Target.AttackLog[Target.AttackLog.first]

		-- Check it against the last time we saw ourself attack
		if entry > atkTime then
			-- Break out of this while loop
			break
		end

		-- Pop it off, since its old
		List.popleft(Target.AttackLog)
	end

	-- Check to see if there are any left
	if List.size(Target.AttackLog) < 1 then
		-- Return 0ms
		return 0,0
	end

	-- Get the time since the most recent attack attempt
	return List.size(Target.AttackLog),GetTick() - atkTime
end

-- Reset Attack Log
function AI_ResetAttackLog()
	Target.AttackLog = List.new()
end

-- Add an attack entry
function AI_AddAttackLog()
	List.pushright(Target.AttackLog,GetTick())
end

-----------------------
-- Support Functions --
-----------------------

-- Determine valid enemies from those targetting a specific actor
function ActorEnemies(id,t)
	-- Our return value
	local enemies

	-- Check if we have an existing table
	if t ~= nil then
		-- Add to the table
		enemies = t
	else
		-- Create a new table
		enemies = Table.new()
	end

	-- Grab the list of actors targetting id
	local actors = ActorTargets[id]

	-- Check if there is a list of actor
	if actors ~= nil then
		-- Loop through each actor
		local i,actor,motion
		for i,actor in ipairs(actors) do
			-- Check their motion
			motion = GetV(V_MOTION,actor)

			-- Check if the actor is in sight and not dead
			if motion ~= MOTION_DEAD and ActorIsInSight(actor) then
				-- Check if the actor is a valid enemy
				if IsValidAttacker(actor) and MonsterAllowed(actor) and
					not KillSteal(actor)
				then
					-- Add it to the list of enemies
					Table.insert(enemies,actor)
				end
			end
		end
	end

	return enemies
end

-- Identify enemies in PVP
function IsValidAttacker(actor)
	-- Our alchemist and ourself are never valid attackers
	if actor == OwnerID or actor == HomunID then
		return false
	end

	-- Check if the actor is a monster
	if IsMonster(actor) ~= 1 then
--[[
	Note: PVP players show up as enemies unless they're
		in your party.

	UPDATE: NOT TRUE!! PVP players always show up as enemies!
--]]
		-- Not a valid enemy
		return false
	end

	-- Check if its a PVP zone
	if not PVPZone then
		return true
	end
	
	-- Check the action thats being used against us
	local motion = GetV(V_MOTION,actor)

	if motion == MOTION_ATTACK or motion == MOTION_ATTACK2 or
		motion == MOTION_CASTING or motion == MOTION_SKILL then
		-- They're probably attacking us

		-- And one last check of who they're targetting
		local target = GetV(V_TARGET,actor)
		if target == HomunID or target == OwnerID or
			IsFriend(target) then

			--[[

		NOTE: If one friend attacks another friend, it will still
			identify them as a valid attacker. This is to allow
			the homunculus to "reprimand" friends for attacking
			other friends

			--]]


			return true
		else
			return false
		end
	else
		-- Not attacking us
		return false
	end

	-- Otherwise return true
	return true
end

-- Check if we would kill steal by attacking the actor
function AI_KillSteal(actor)
--[[
	-- Check if we're in a PVP zone
	if PVPZone then
		-- Can't kill-steal in a PVP zone

		-- And never check again
		KillSteal = FalseReturn

		return false
	end
--]]

	-- Check if the monster is free for all
	if MIgnoreKS(actor) then
		-- Free-for-all monsters are never Kill-Stealed
		return false
	end

	-- Check if its a monster
	if IsMonster(actor) ~= 1 then
		-- Can't kill-steal a player O.o
		return false
	end

	-- Check if the monster is targetted and locked
	if actor == Target.ID and Target.Lock then
		-- It might be kill stealing, but we won't count it
		return false
	end

	-- Get the monster's target
	local target = GetV(V_TARGET,actor)

	-- Check if the monster is targetting something
	if target ~= 0 then
		-- Check if the monster is targetting homunculus or owner
		if target == HomunID or target == OwnerID then
			-- Won't be "kill-stealing"
			return false
		end

		-- Check if monster is targetting a friend
		if DefendFriends and IsFriend(target) then
			-- We are supposed to defend friends,
			-- so it isn't kill-stealing
			return false
		end

		--[[ Check if monster is targetting another monster
			(with a heal spell or such)

			NOTE: If the monster is targetting another monster, we will
				procede to check if its being targetted by anybody else
		--]]
		if IsMonster(target) ~= 1 then

			-- Check if the target is in sight range
			if AABlocks(OwnerID,target) <= SightRange + 2 then
				-- It would be kill stealing
				return true
			end
		end
	end

	-- Check if any non-monsters are targetting it
	local table = ActorTargets[actor]
	if table ~= nil then
		if Table.getn(table) > 0 then
			-- Loop through, trying to find a non-monster
			local i,actor
			for i,actor in ipairs(table) do
				if actor ~= HomunID and
					actor ~= OwnerID and
					IsMonster(actor) ~= 1 and
					not IsFriend(actor) and
					(AABlocks(OwnerID,actor) <= SightRange + 2) then

					-- Its not a monster or a friend,
					-- so likely we'd be kill stealing
					-- (or its outside of sight)
					return true
				end
			end
		end
	end

	-- It won't kill steal
	return false
end
