--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

----------------------------
-- Main Command Processor --
----------------------------

-- Default command processing
function AI_Command(command,shiftkey)
	-- Set the time of the last command
	OwnerLast.Command = GetTick()

	-- Sequence item to be added
	local seq = nil

	-- Check for special-case commands
	if command[1] == FOLLOW_CMD then
		-- Pretend shift-key was down, so we don't
		-- reset anything
		shiftkey = true
	end

	-- Reset holding position
	HoldPosition = false

	-- Check if we're clearing the sequencer
	if not shiftkey
		and command[1] ~= SKILL_OBJECT_CMD
		then
		-- Clear our sequencer queue
		List.clear(AI_seq)

		-- Reset the target lock
		Target.Lock = false

		-- Stop any existing movement
		if GetV(V_MOTION,HomunID) == MOTION_MOVE then
			Move(HomunID,GetV(V_POSITION,HomunID))
		end
	end

	-- Decide what to do, based on the command
	if command[1] == MOVE_CMD then
		-- "alt+right click" on ground

		CMD_AdvMove(command,shiftkey)

	elseif command[1] == ATTACK_OBJECT_CMD then
		-- "alt+right click" on monster, twice

		CMD_AdvAttack(command,shiftkey)

	elseif command[1] == SKILL_OBJECT_CMD then
		-- Skill targetted on enemy

		CMD_SkillObj(command,shiftkey)

	elseif command[1] == FOLLOW_CMD then
		-- "Alt+T"

		-- Alternate modes
		if Mode == MODE_AGGRO then
			Mode = MODE_PASSIVE
			Log(0,"Mode changed to: Passive")
		else
			Mode = MODE_AGGRO
			Log(0,"Mode changed to: Aggressive")
		end

	else
		-- Unknown command

		local msg = string.format("Unknown GetMsg() command: %d(",
			command[1])

		local i
		for i=2,5 do
			if command[i] == nil then
				break
			end

			if i ~= 2 then
				msg = msg .. ", "
			end
			msg = string.format("%s%d",msg,command[i])
		end
		msg = msg .. ")"
		Log(0,msg)
	end
end

---------------------------------
-- Command Processor Functions --
---------------------------------

-- Function to determine if a Quickdial command was used on our owner
function CMD_AdvMove_QD(cmd,shiftkey)
	-- Get the x,y coords
	local x,y = cmd[2],cmd[3]

	-- Check the distance of the move from our owner
	if APBlocks(OwnerID,x,y) < 2 then
		-- Get the owner's X and Y position
		local ax,ay = GetV(V_POSITION,OwnerID)

		-- Check if we should do the move command
		if AMC_PreParse(OwnerID,x-ax,y-ay) and not shiftkey then
			-- Move command should continue
			CMD_Move(cmd)
		end

		return
	end

	-- Check if its a quickdial thats targetting a person that isn't our owner
	return CMD_AdvMove_AMC(cmd,shiftkey)
end

-- Function to determine if an AMC command was used
function CMD_AdvMove_AMC(cmd,shiftkey)
	-- Get the x,y coords
	local x,y = cmd[2],cmd[3]

	-- Loop through all the actors, making a list of those that are close enough
	local actors = Table.new()
	local i,actor,motion,dist
	for actor,i in pairs(ActorTargets) do
		-- Check that the actor isn't the homunculus
		if actor ~= HomunID then
			-- Check if the actor is close enough to
			-- be the target of an AMC
			if APBlocks(actor,x,y) < 2 then
				-- Add the actor to the list
				Table.insert(actors,actor)
			end
		end
	end
	
	-- Set the actor to -1
	actor = -1
	
	-- Check if we have an AMC target
	if AMCTarget ~= -1 then
		-- Check if the target is in our actors
		if ActorIsInTable(AMCTarget,actors) then
			-- It is, so its our actor this time around
			actor = AMCTarget
		end
	end

	-- Check if we didn't find an actor yet
	if actor == -1 then
		-- Get the nearest actor
		actor = PTNearest(x,y,actors)
	end
	
	-- Check if there is no closest actor
	if actor == -1 then
		-- Its just a simple move command

		CMD_Move(cmd)		
		
		return
	end

	-- Get actor's x,y coords
	local ax,ay = GetV(V_POSITION,actor)
	
	-- Check if we should do the move command
	if AMC_PreParse(actor,x-ax,y-ay) and not shiftkey then
		-- Move command should continue
		CMD_Move(cmd)
	end
end

-- Function to determine if a move command was an attack
function CMD_AdvMove_Atk(cmd,shiftkey)
	-- Get the x,y coords
	local x,y = cmd[2],cmd[3]
	
	-- Loop through all the actors, making a list of those that are close enough
	local actors = Table.new()
	local i,actor,motion,dist
	for actor,i in pairs(ActorTargets) do
		-- Check that the actor isn't the homunculus or our owner
		if type(actor) == "number" and actor ~= HomunID and actor ~= OwnerID then
			-- Check if the actor is dead
			if GetV(V_MOTION,actor) ~= MOTION_DEAD then
				-- Get the actor's distance
				dist = APDistance(actor,x,y)

				-- Check if the actor is a monster and its close enough
				if IsMonster(actor) == 1 and
					dist <= MoveIsAttackWhenDistance then

					-- Add the actor to the list
					Table.insert(actors,actor)
				end
			end
		end
	end
	
	-- Set the actor to -1
	actor = -1
	
	-- Prioritize the table
	actors = MPrioritySieve(actors)
	
	-- Get the nearest actor
	actor = PTNearest(x,y,actors)

	-- Check if there is no closest actor
	if actor == -1 then
		-- Its just a simple move command

		CMD_Move(cmd)		
		
		return
	end

	-- Get actor's x,y coords
	local ax,ay = GetV(V_POSITION,actor)
	
	-- Log it
	Log(0,"Interpretting move (%d,%d) as attack monster #%d at (%d,%d).",
		x,y,actor,ax,ay)

	-- Check if we're supposed to kite this monster
	if not MKite(actor) then
		-- Issue the attack command
		CMD_Attack({ SEQ_ATTACK, actor })
	else
		-- Issue the kite command
		CMD_Kite({ SEQ_KITE, actor })
	end
end

-- Simple move command
function CMD_Move(command)
	-- Queue a command to set HoldPosition
	List.pushright(AI_seq, { SEQ_SET_HOLD, true })

	-- Queue the move command
	List.pushright(AI_seq,command)
end

-- Attack command, with possible AMC
function CMD_AdvAttack(command,shiftkey)
	-- Check the target
	local target = command[2]
	
	-- Check if the target is the same as our AMCTarget
	if target == AMCTarget then
		-- Check if theres a sequence
		if AMCSequence ~= "" then
			-- Execute the sequence
			local ret = AMC_Execute(AMCTarget,AMCSequence)
			
			-- Clear the sequence, and reset the timer
			AMCSequence=""
			AMCLastTime=0
			
			-- Check if we should continue with the
			-- attack command
			if not ret or shiftkey then
				-- We shouldn't
				return
			end
		end
	end
	
	-- Set the AMCTarget
	AMCTarget = target

	-- Check if we're supposed to kite this monster
	if not MKite(command[2]) then
		-- Use the attack command
		CMD_Attack(command)
	else
		-- Use the kite command
		CMD_Kite(command)
	end
end

-- Attack command
function CMD_Attack(command)
	-- Queue a command to set the target
	List.pushright(AI_seq, { SEQ_SET_TARGET, command[2], true } )

	-- Add the command to the sequencer queue
	List.pushright(AI_seq,command)
end

-- Kite command
function CMD_Kite(command)
	-- Make sure command[1] is SEQ_KITE (it'll probably usually be SEQ_ATTACK)
	command[1] = SEQ_KITE

	-- Queue a command to set the target
	List.pushright(AI_seq, { SEQ_SET_TARGET, command[2], true } )

	-- Add the command to the sequencer queue
	List.pushright(AI_seq, command)
end

-- Skill command
function CMD_SkillObj_Original(command)
	-- Since they can't shift click for a skill, we'll add it to the front
	List.pushleft(AI_seq,command)
end

-- Skill command, which triggers attack command
function CMD_SkillObjAndAttack(command)
	-- Use the skill first
	List.pushright(AI_seq,command)
	
	-- Attack after using the skill
	CMD_Attack( { SEQ_ATTACK, command[4] } )
end

--------------------------------
-- Advanced Movement Commands --
--------------------------------

-- Initialize AMCs
function AMC_Init()
	-- Check if Advanced Movement Commands are disabled
	if not EnableAMC then
		-- They are disabled, so remove all the functions
		AMC_Init = NoReturn
		AMC_Check = NoReturn
		AMC_PreParse = TrueReturn
		AMC_Parse = TrueReturn
		AMC_Execute = TrueReturn
		AMC_Reset = NoReturn
		CMD_AdvMove = CMD_AdvMove_Atk
		CMD_AdvAttack = CMD_Attack
		return
	end

	-- Check if we're using quickdial AMCs
	if QuickdialAMC then
		-- Using quickdial AMCs

		-- Use the quickdial parser
		AMC_Parse = AMC_Parse_QD

		-- Set our advanced move command
		CMD_AdvMove = CMD_AdvMove_QD

		-- Use the regular attack with quickdials
		CMD_AdvAttack = CMD_Attack

		-- Quickdial AMCs don't ever timeout
		AMC_Check = NoReturn
	else
		-- Not using quickdial AMCs

		-- Use the regular parser
		AMC_Parse = AMC_Parse_AMC

		-- Set our advanced move command
		CMD_AdvMove = CMD_AdvMove_AMC
	
	end
end

-- Check when the last received command was
function AMC_Check()
	-- Check if we should be checking... XD
	if AMCLastTime == 0 then
		return
	end
	
	-- Get the current tick count
	local ticks = GetTick()
	
	-- Check if its time to clear
	if ticks >= AMCLastTime + AMCResetTime then
		-- Log it
		Log(0,"AMC sequence timed out.")
		
		-- Clear everything
		AMC_Reset()
	end
end

-- Function to determine the subcommand for an AMC
function AMC_PreParse(target,x,y)
	-- Determine the subcommand "name", based on the x- and y-deltas
	
	-- Thanks for Fallen~Angel for the following bit of code
	local subcmd = (5+x) + (y*3)
	
	if subcmd < 1 or subcmd > 9 then
		-- Invalid, but this should never happen

		-- Grab the target's x and y
		local ax,ay = GetV(V_POSITION,target)
		
		-- Log it
		Log(0,"Invalid quickdial command (%d,%d). Resetting AMC sequence.",x-ax,y-ay)
		
		-- Reset
		AMC_Reset()
		
		-- Continue with the move command
		return true
	end
	
	-- Do the actual parsing
	return AMC_Parse(target,subcmd)
end

-- Function to reset the Advanced Movement Commands
function AMC_Reset()
	AMCLastTime = 0
	AMCSequence = ""
	AMCTarget = -1
end

-- Function to add a full command to the AMCs
function AMC_AddCommand(cmd,f)
	AMC_DelCommand(cmd)
	Table.insert(AMCFullCommands,{ cmd, f })
end

-- Function to add a subparser to the AMCs
function AMC_AddParser(cmd,f)
	AMC_DelCommand(cmd)
	Table.insert(AMCSubParsers,{ cmd, f })
end

-- Function to remove a parser or command that has
--	a sequence of "cmd"
function AMC_DelCommand(cmd)
	-- Loop through each command
	local i,max = 0,Table.getn(AMCFullCommands)
	for i=1,max do
		-- Check the command
		if AMCFullCommands[i][1] == cmd then
			-- Remove it
			Table.remove(AMCFullCommands,i)

			-- Decrement max
			max = max - 1
		end
	end

	-- Loop through each subparser
	max = Table.getn(AMCSubParsers)
	for i=1,Table.getn(AMCSubParsers) do
		-- Check the command
		if AMCFullCommands[i][1] == cmd then
			-- Remove it
			Table.remove(AMCFullCommands,i)

			-- Decrement max
			max = max - 1
		end
	end
end

---------------------------------
-- Quickdial Parse and Execute --
---------------------------------

function AMC_Parse_QD(target,subcmd)
	-- Execute the command, and let it decide if we should move there
	return AMC_Execute_QD(target,tostring(subcmd))
end

function AMC_Execute_QD(target,command,reused)
	-- Loop through each of the full command executers
	local i,max = 1,Table.getn(AMCFullCommands)
	for i=1,max do

		-- Check if the executer is the one we need
		if command == AMCFullCommands[i][1] then
			-- It is
			
			-- Log it
			Log(0,"Executing AMC Command: %s (%d)",
				command,target)
			
			-- Execute it
			local func = AMCFullCommands[i][2]
			return func(target),true
		end
	end
	
	if not reused then
		-- Its an unrecognized Quickdial command
		Log(0,"Unrecognized QuickDial Command: %q.",command)
	end

	return true,false
end

-------------------------------------
-- Non-quickdial Parse and Execute --
-------------------------------------

function AMC_Parse_AMC(target,subcmd)
	-- Set the last time
	AMCLastTime = GetTick()
	
	-- Check if we have a new target
	if target ~= AMCTarget then
		-- Log it
		Log(0,"Setting AMC target as Actor #%d (previously #%d)",target,AMCTarget)
		
		-- Set the target
		AMCTarget = target
		
		-- Reset the sequence
		AMCSequence = ""
	end

	-- Check if its the special subcommand "5", to execute
	if subcmd == 5 then
		-- Execute the sequence
		local ret = AMC_Execute_AMC(AMCTarget,AMCSequence)
		
		-- Clear the sequence, and reset the timer
		AMCSequence=""
		AMCLastTime=0
		
		-- Continue with the move if the function wants us to
		return ret
	end
	
	-- Otherwise concatenate the subcmd to the sequence
		-- (ensure they're not converted to numbers)
	AMCSequence = string.format("%s%d",AMCSequence,subcmd)
	
	-- Continue with move command
	return true
end

function AMC_Execute_AMC(target,command)
	-- Check if the command is empty
	if command == "" then
		-- Do nothing, but move to them
		return true
	end
	
	-- Reuse the quickdial execute
	local ret,found = AMC_Execute_QD(target,command,true)

	-- Check if we found it
	if found then
		-- Return the AMC commands desired return
		return ret
	end
	
	-- We're not using quickdial, so it can be a subparser

	-- Loop through each of the sub parsers
	local i,max = 1,Table.getn(AMCSubParsers)
	for i=1,max do
		-- Get the subparser's beginning command
		local subcmd = AMCSubParsers[i][1]

		-- Check if its the right one
		if string.sub(command,1,string.len(subcmd)) == subcmd then
			-- Its a subparser

			-- Get the argument
			local arg = string.sub(command,string.len(subcmd)+1)

			-- Log it
			Log(0,"Executing AMC sub-parser: %s (%d,%s)",
				subcmd,target,arg)

			-- Parse it
			local func = AMCSubParsers[i][2]
			return func(target,arg)
		end
	end
	
	-- Log an Unrecognized AMC command
	Log(0,"Unrecognized Advanced Movement Command: %q.",command)
	
	return true
end
