--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

-----------------------
-- Filir-specific AI --
-----------------------

--[[
	Unsupported Skills:

	Accelerated Flight	(ID 8011, SP { 30, 40, 50, 60, 70 } )
	S.B.R. 44		(ID 8012, SP ? )
--]]

-- SP values
FlittingSP = { 30, 40, 50, 60, 70 }
AccFlightSP = { 30, 40, 50, 60, 70 }

-- Delays
FlittingDelay = { 60, 70, 80, 90, 120 }

-- Durations
FlittingDuration = { 60, 55, 50, 45, 40 }

-- Filir-specific monster database
FilirMDB = { }

-- Attempt variables
FlittingAttempted = 0
FlittingAttemptSP = 0
FlittingAttemptTicks = 0

-- The timeout for the next skill
FlittingTimeout = 0
FlittingSPTimeout = 0

-- Initialization function
function Filir_Init()
	-- Add the Moonlight skill to the attack skill list
	AddAttackSkill(8009, { 4, 8, 12, 16, 20 },2)

	-- Check if support skills are enabled
	if SupportSkills then
		-- Set the skill check
		HomunSkillCheck = Filir_CheckSkills

		-- Hook the SetTarget function
		PreFilir_SetTarget = SetTarget
		SetTarget = Filir_SetTarget

		-- Set the -1 type (idea from Truly Lazy)
		FilirMDB[-1] = { ticks=0, targets=0, avgTicks=0 }

		-- See if we should use the duration or the delay
		if FilirUseDuration then
			-- Override the delay with the duration
			FlittingDelay = FlittingDuration
		end

		-- See if we're supposed to use Flitting
		if FlittingLevel > 0 then
			-- Add the flitting check
			CheckFlitting = Filir_CheckFlitting
		else
			-- Empty flitting check, so we don't use it
			CheckFlitting = FalseReturn
		end

		-- See if we're supposed to use Accelerated Flight
		if false then
			CheckAccFlight = Filir_CheckAccFlight
		else
			CheckAccFlight = FalseReturn
		end
	end
end

-- Function to set the target
function Filir_SetTarget(id)
	-- Make sure the id is a number
	if type(id) ~= "number" then
		return PreFilir_SetTarget(id)
	end

	-- Get the type of the monster
	local type = GetV(V_HOMUNTYPE,id)

	-- Check to see if we've encountered this type before
	if FilirMDB[type] == nil then
		-- We haven't, make a table for it
		FilirMDB[type] = { ticks=0, targets=0, avgTicks=0 }
	else
		-- Check if the ID is different from our current target
		if id ~= Target.ID and type ~= -1 then
			-- Get the time it was targetted
			local ticks = GetTick() - Target.StartTick

			-- Check if the time it was targetted was longer than .25 seconds
			if ticks > 250 then
				-- Add it to our table
				FilirMDB[type].ticks = FilirMDB[type].ticks + ticks
				FilirMDB[type].targets = FilirMDB[type].targets + 1

				-- Calculate the new average ticks
				FilirMDB[type].avgTicks =
					FilirMDB[type].ticks / FilirMDB[type].targets
			end
		end
	end

	-- Call the original set target
	PreFilir_SetTarget(id)
end

-- Function to check if we should use either flitting or accelerated flight
function Filir_CheckSkills()
	-- Check if we want to use accelerated flight
	if not CheckAccFlight() then
		-- We don't, so check if we want to use flitting
		return CheckFlitting()
	end

	-- We want to use accelerated flight, stop the attack skills
	return true
end

-- Function to check if we should use Flitting
function Filir_CheckFlitting()
	-- Default to returning false
	local ret = false

	-- Get our usable amount of SP
	local sp = SPUsable(HomunID)

	-- Get the current time
	local ticks = GetTick()

	-- Check if we are allowed to use a skill against our current target
	local aSkillID,aSkillLevel = CheckAttackSkill(Target.ID)

	-- Check if its moonlight
	if aSkillID == 8009 then
		-- Subtract the amount of SP required to cast it from
		--	our current available SP
		sp = sp - (aSkillLevel * 4)
	end

	-- See if we need to be saving SP and we don't have enough
	if sp <= FlittingSP[FlittingLevel] and FlittingSPTimeout <= ticks then
		-- We're saving SP
		ret = true
	end

	-- Undo any previous available sp modification
	if aSkillID == 8009 then
		sp = sp + (aSkillLevel * 4)
	end

	-- Check if we can use the skill again yet
	if ticks < FlittingTimeout then
		-- Can't use the skill yet
		return ret
	end

	-- Check if the owner is sitting down
	if OwnerMotion == MOTION_SIT then
		-- Can't do anything yet
		return ret
	end

	-- Check if we already tried to cast Flitting
	if FlittingAttempted ~= 0 then
		-- Check if it was successful
		if sp < FlittingAttemptSP then
			-- Determine the level that was used
			local i,level
			for i=1,FlittingAttempted do
				level=i
				-- Check if we used less than this level's SP cost
				if FlittingAttemptSP - sp <= FlittingSP[i] then
					-- Assume we used this level
					break
				end
			end

			-- Log it
			Log(2,"Used flitting (level %d) to increase killing speed.",level)

			-- Set the timeout for flitting
			FlittingTimeout = FlittingAttemptTicks + (FlittingDelay[level] * 1000)

			-- Set the timeout for saving SP
			FlittingSPTimeout = FlittingAttemptTicks + FlittingSPSaveTicks

			-- Reset the attempt variables
			FlittingAttempted = 0

			-- Return a recurse
			return Filir_CheckFlitting()

		-- Check if it timed out?
		elseif FlittingAttemptTicks + 300 < ticks then
			-- Flitting failed

			-- Log it
			Log(2,"Flitting apparently failed.")

			-- Reset the attempt variables
			FlittingAttempted = 0

			-- Return a recurse
			return Filir_CheckFlitting()
		end

		-- If it didn't time out, and it wasn't successful, just wait for it
		return true
	end

	-- Check if we're targetting something
	if Target.ID == -1 then
		-- We aren't, so we don't want to use flitting yet
		return ret
	end

	-- Check if we're locked on
	if not Target.Lock then
		-- Don't use flitting yet
		return ret
	end

	-- Get our current target's type
	local type = GetV(V_HOMUNTYPE,Target.ID)

	-- Check if its long enough
	if FilirMDB[type].avgTicks < FlittingMinTicks then
		-- Not long enough, so we don't want to use flitting
		return ret
	end

	-- Check if we have enough SP
	if sp >= FlittingSP[FlittingLevel] then
		-- Log it
		Log(2,"Attempting to use Flitting (level %d), to kill actor #%d (type #%d) faster.",
			FlittingLevel,Target.ID,type)

		-- Cast Flitting
		SkillObject(HomunID, FlittingLevel, 8010, HomunID)

		-- Set the attempt variables
		FlittingAttempted = FlittingLevel
		FlittingAttemptSP = sp
		FlittingAttemptTicks = ticks
	end

	-- We're trying to use flitting
	return true
end

-- Function to check if we should use accelerated flight
function Filir_CheckAccFlight()

	-- NOTE: This will be supported as soon as it works in RO.

	if true then
		-- We don't want to use something that doesn't work
		return false
	end

	-- Cast Accelerated Flight (never)
	SkillObject(HomunID, 5, 8011, HomunID)
end