--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

---------------------------
-- Monster Table Support --
---------------------------

-- Convert the MonsterDB table
function ConvertMDB()
	-- Log it
	Log(0,"Optimizing monster database...")
	
	-- Verify we have a valid default
	if type(DefaultMonsterSettings) == "table" then
		-- Make sure the "ID" will not be counted as the end of the table
		DefaultMonsterSettings[1] = 0
	else
		-- Generate a blank table
		DefaultMonsterSettings = { }
	end

	-- Verify our default settings
	DefaultMonsterSettings = VerifyMonSettings(DefaultMonsterSettings,true,"default")

	-- Make a copy of the monster db
	local mdb = { }

	-- Loop through each monster type
	local i,t
	for i,t in pairs(MonsterDB) do
		-- Grab the ID
		local id = t[1]

		-- Verify the ID
		if type(id) == "number" then
			-- Add it to our new database
			--mdb[id] = VerifyMonSettings(t,false,tostring(id))

			-- Add it to our new database
			mdb[id] = t

			-- Make sure "verified" is false
			mdb[id].verified = false
		end
	end

	MonsterDB = mdb
	
	-- Ensure this isn't repeated
	ConvertMDB = NoReturn
end

-- Verify monster settings
function VerifyMonSettings(t,isDefault,id)
	-- Check if we're working on the default
	local def
	if isDefault then
		def = { 0, 1, true, 5, 2000, 1, 3, 3000, -1, false, false, false }
	else
		def = DefaultMonsterSettings
	end

	-- Check if they have a high enough level
	if DebugLevel >= 8 then
		-- Show them what the monster entry looked like before
		local j,k,str = 0,0,string.format("Before: MDB[%s] = { ",id)
		for j,k in pairs(t) do
			if j ~= "verified" then
				str = str .. string.format("%s=%q, ",
					tostring(j), tostring(k))
			end
		end
		Log(8,"%s },",str)
	end

	-- Make the table of types
	--local types = { }
	local i,val
	for i,val in pairs(def) do
		-- Get the type of value, and set it in our types
		--types[i] = type(val)

		-- Check if theres a value in our table
		if t[i] == nil and i ~= "verified" and i ~= "n" then
			-- Log it
			Log(7,"Missing option %q for monster ID %q",
				tostring(i),id)

			-- Set the option
			t[i] = val
		end
	end

	-- Loop through each value of the monster table
	for i,val in pairs(t) do
		--[[

			NOTE: This section of code is disabled to better allow for
				Custom mods to add Monsters.lua settings. It is recommended
				that any options Custom mods use be named options:

		example:	{ 123, 4, true, 5, 6, 7, 8, 9, customOption1 = "hi" },

		-- Check if theres an option in our types
		if types[i] == nil then
			-- Log it
			Log(5,"Monster ID %q has unknown option: %q",
				id,tostring(i))

			-- No type, so remove the option
			t[i] = nil
		else

		--]]

			-- Check if theres a default type for this option
			if def[i] ~= nil then

				-- Theres an option, check its type
				if type(val) ~= type(def[i]) then
					-- Log it
					Log(2,"Invalid option %q for monster ID %q",
						tostring(i),id)

					-- Wrong type, so set it to our default
					t[i] = def[i]
				end

			end

		-- end
	end

	-- Check if they have a high enough level
	if DebugLevel >= 8 then
		-- Show them what the monster entry looks like afterwards
		local j,k,str = 0,0,string.format("After: MDB[%s] = { ",id)
		for j,k in pairs(t) do
			if j ~= "verified" then
				str = str .. string.format("%s=%q, ",
					tostring(j), tostring(k))
			end
		end
		Log(8,"%s },",str)
	end

	-- This table was verified
	t.verified = true

	-- Return our verified table
	return t
end

-- Get the monster-specific table
function MDBTable(actor)
	-- Make sure the actor isn't nil
	if actor ~= nil then
		-- Get the actor type
		local type = GetV(V_HOMUNTYPE,actor)

		-- Get the table
		local t = MonsterDB[type]

		-- Check if this is monster-specific
		if t ~= nil then
			-- It is, return the monster-specific table

			-- Check if its been verified yet
			if not t.verified then
				-- Log it
				Log(2,"First access of monster settings for Monster ID #%d, verifying...",
					type)

				-- Verify it
				MonsterDB[type] = VerifyMonSettings(t,false,tostring(type))

				-- Regrab the table (in case it changes for some reason)
				t = MonsterDB[type]
			end

			-- Return the table
			return t
		end
	end

	-- Its not monster-specific, return the default
	return DefaultMonsterSettings
end

-------------------------------
-- Monster Selection Support --
-------------------------------

-- Check if we're allowed to target the monster
function MonsterAllowed(actor)
	-- Check if either skills or melee are allowed against the monster 
	if MSkillsAllowed(actor) or MAttackAllowed(actor) then
		return true
	end
	
	return false
end

-- Check if we are allowed to melee the monster
function MAttackAllowed(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Return the melee allowed value
	return t[3]
end

-- Check if the monster is free-for-all, and can be KSed at any time
--	** idea from HauNix of RagnaInfo forums **
function MIgnoreKS(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Return the value
	return t[12]
end

-- Check if we're supposed to target this monster only if it attacks us
function MNoAggroAgainst(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Return the defend only value
	return t[10]
end

-- Check if we're supposed to kite this monster rather than attack it
function MKite(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Return the kite value
	return t[11]
end

-- Function to remove all the monsters that are supposed to be kited,
--	unless they're all kite monsters
function MKiteSieve(t)
	local max = Table.getn(t)

	-- Check if we have any
	if max == 0 then
		-- Return a new table
		return Table.new()
	end

	-- Setup the default table
	local tablekite = true
	local monsters = Table.new()

	local i,actor,monkite
	for i=1,max do
		-- Get the actor
		actor = t[i]

		-- Check if the monster should be kited
		monkite = MKite(actor)

		-- Check if our table is currently full of kitable monsters
		if tablekite and not monkite then
			-- We are no longer filling with kitable monsters
			tablekite = false

			-- Make a new table
			monsters = Table.new()
		end

		-- Check if we should add to the table
		if tablekite == monkite then
			-- Add to the table
			Table.insert(monsters,actor)
		end
	end

	-- Return the table of monsters
	return monsters
end

-- Check the priority level of a monster
function MAttackPriority(actor)
	-- Check the mosnter's type
	local t = MDBTable(actor)
	
	-- Return the priority level
	return t[2]
end

-- Function to select the highest priority monsters from a list
function MPrioritySieve(t)
	-- Get the highest value of the table
	local max = Table.getn(t)
	
	-- Check if we have none
	if max == 0 then
		return Table.new()
	end

	-- Setup the default table
	local priority = MAttackPriority(t[1])
	local monsters = Table.new()
	
	-- Loop through our table, sorting the highest priority monsters into our new table
	local i,actor,prio
	for i=1,max do
		-- Get the actor
		actor = t[i]
		
		-- Check the actor's priority
		prio = MAttackPriority(actor)
		
		-- Check if we should start a new priority table
		if prio > priority then
			-- Priority is higher, start a new table
			priority = prio
			monsters = Table.new()
		end
		
		-- Check if we should add the monster to the table
		if prio == priority then
			-- Add the monster
			Table.insert(monsters,actor)
		end
	end
	
	-- Return the table with the highest priority monsters
	return monsters
end

------------------------------
-- Homunculus Skill Support --
------------------------------

-- Check if we are allowed to use skills against the monster
function AI_MSkillsAllowed(actor)
	-- Check if we're pausing skills for the moment
	if PauseAttackSkill then
		-- They're paused
		return 0
	end

	-- Get the monster's type
	local t = MDBTable(actor)

	-- Check if level and max casts are greater than 0
	if t[4] > 0 and t[6] ~= 0 then
		return true
	end

	-- If the value isn't greater than 0, disable skill usage
	return false
end

-- Check the monster-specific skill timeout
function MSkillTimeout(actor)
	-- Get the monster's type
	local t = MDBTable(actor)
	
	-- Return the timeout value
	return t[5]
end

-- Get the maximum level of an attack skill to use against a monster
function MSkillMaxLevel(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Return the max skill level value
	return t[4]
end

-- Get the maximum number of times to cast a homunculus skill against the monster
function MSkillMaxCasts(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Get the max number of casts
	local ret = t[6]

	-- Check if it is negative
	if ret < 0 then
		--[[
			Return 1 higher than our current number of casts
			against our target (so that its seemingly infinite)
		--]]

		ret = Target.Skill + 1
	end

	-- Return the max number of casts
	return ret
end

-----------------------------
-- Auto-Sword Bolt Support --
-----------------------------

-- Check if we are allowed to sword-bolt the monster
function MSwBoltAllowed(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Check if sword-bolts are allowed
	if t[7] > 0 and t[9] ~= 0 then
		return true
	end

	-- They aren't
	return false
end

-- Get the timeout against this monster
function MSwBoltTimeout(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Return the timeout
	return t[8]
end

-- Get the max level of bolt to cast
function MSwBoltMaxLevel(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Return the level
	return t[7]
end

-- Get the max number of casts against the monster
function MSwBoltMaxCasts(actor)
	-- Get the monster's type
	local t = MDBTable(actor)

	-- Get the max number of casts
	local ret = t[9]

	-- Check if its negative
	if ret < 0 then
		--[[
			Negative means infinite, so add 1 to our
			current count of casts
		--]]

		ret = AlchTarget.Skill + 1
	end

	return ret
end


----------------------------
-- Monster Ignore Support --
----------------------------

-- Remove all the ignored monsters from a Table
function IgnoredSieve(t)
	-- Loop through each monster in the table
	local i,max = 1,Table.getn(t)
	while i <= max do
		-- Check if the monster is ignored
		if IsIgnored(t[i]) then
			-- Monster is ignored, remove it
			Table.remove(t,i)

			-- Reduce the max
			max = Table.getn(t)
		else
			-- Monster isn't ignored, move to next
			i = i + 1
		end
	end

	return t
end

-- Check our list of ignored monsters
function CheckIgnored(ticks)
	-- Loop through all of our ignored monsters
	local actor,time
	for actor,time in pairs(IgnoredTargets) do
		-- Check if its ignored
		if IgnoredTargets[actor] > 0 then
			-- Subtract the time from the target
			IgnoredTargets[actor] = IgnoredTargets[actor] - ticks
	
			-- Log it
			Log(3,"Removed %d ticks from Actor #%s's ignore time, resulting in %d",
				ticks, actor, IgnoredTargets[actor])
		end
	end
end

-- Check if a monster is ignored
function IsIgnored(actor)
	-- Loop through our ignored monsters
	if type(IgnoredTargets[actor]) == "number" and
		IgnoredTargets[actor] > 0 then
		-- They still have time left
		return true
	end

	-- If its not in the ignored list, its not ignored
	return false
end

-- Add a monster to the ignore list
function AddIgnore(actor,time)
	-- Check if its already ignored
	if IsIgnored(actor) then
		-- Do nothing
		return
	end

	-- Check the time
	if type(time) ~= "number" then
		time = ActorIgnoreTime
	end
	
	-- Log it
	Log(0,"Ignoring actor #%d for %d milliseconds.",actor,time)

	-- Add the actor to the ignore list
	IgnoredTargets[actor] = time
end