--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

------------------------
-- Sequence Processor --
------------------------

function ProcessSequence()
	-- Loop until a sequence doesn't tell us to keep going
	local cont,seq,func
	repeat
		-- Make sure there are sequences in our stack
		if List.size(AI_seq) < 1 then
			Log(10,"SEQ: 0 { nil }")

			break
		end

		-- Grab the sequence
		seq = List.popleft(AI_seq)

		-- Determine which function to call
		if seq[1] == SEQ_MOVE then
			func = SEQ_PMove
		elseif seq[1] == SEQ_ATTACK then
			func = SEQ_Attack
		elseif seq[1] == SEQ_SKILL_OBJECT then
			func = SEQ_ASkill
		elseif seq[1] == SEQ_SKILL_AREA then
			func = SEQ_PSkill
		elseif seq[1] == SEQ_MOVE_TO then
			func = SEQ_AMove
		elseif seq[1] == SEQ_MOVE_TO_POS then
			func = SEQ_AMove_Pos
		elseif seq[1] == SEQ_MOVE_TO_NEG then
			func = SEQ_AMove_Neg
		elseif seq[1] == SEQ_PREROUTED then
			func = SEQ_PreRouted
		elseif seq[1] == SEQ_KITE then
			func = SEQ_Kite
		elseif seq[1] == SEQ_SET_TARGET then
			-- Set the target
			SetTarget(seq[2])
			Target.Lock = seq[3]

			-- The function will return true to move on
			-- to the next target
			func = TrueReturn
		elseif seq[1] == SEQ_SET_HOLD then
			-- Set the HoldPosition variable
			HoldPosition = seq[2]

			-- Return true to continue sequencing
			func = TrueReturn
		else
			-- Check if the sequencer command is a function
			if type(seq[1]) == "function" then
				func = seq[1]
			else
				func = SEQ_Custom
			end
		end

		if DebugLevel >= 10 then
			local str = string.format("SEQ: %d (",seq[1])
			local i,j
			for i,j in seq do
				if i ~= 1 then
					str = str .. string.format("%s=%q, ",i,tostring(j))
				end
			end
			str = str .. ")"
			Log(10,str)
		end

		-- Ensure the sequence has a log level
		if not seq.loglevel then
			seq.loglevel = 1
		end

		-- Call the function
		cont,seq,seq2 = func(seq)

		-- If a sequence is given back, push it onto the stack
		if seq~= nil then
			-- If 2 are given back, put the 2nd of the stack
			-- first, so its executed 2nd
			if seq2 ~= nil then
				List.pushleft(AI_seq,seq2)
			end

			-- Put the 1st sequence on the stack
			List.pushleft(AI_seq,seq)
		end
	until cont ~= true

	-- Return the next sequence
	if List.size(AI_seq) > 0 then
		return AI_seq[AI_seq.first]
	end

	-- No sequences, return a blank sequence
	return { SEQ_NONE }
end

---------------------------------
-- Sequencer Support Functions --
---------------------------------

-- Check if a specific sequence is in the sequencer que
function InSequence(type)
	-- Check the size of AI_seq
	if List.size(AI_seq) < 1 then
		return false
	end

	-- Loop through AI_seq
	for i=AI_seq.first,AI_seq.last do
		-- Check the sequence type
		if AI_seq[i][1] == type then
			return true
		end
	end

	-- Default to false
	return false
end

-------------------------------
-- Default Sequence Handlers --
-------------------------------

-- Move to a position
function SEQ_PMove(seq)
	-- Grab destination
	local destx,desty = seq[2],seq[3]

	-- Check if we're locked within a certain distance of our owner
	if MaxDistance > 0 then
		-- Check if moving to our destination should be ignored
		if APDistance(OwnerID, destx, desty) > MaxDistance then
			-- Move on to the next sequence
			return true
		end
	end

	-- Grab current position
	local x,y = GetV(V_POSITION,HomunID)

	-- Check if we're moving
	local moving = false
	if HomunMotion == MOTION_MOVE then
		moving = true
	end
	
	-- Check if the destination is out of sight or range
	if not IsInSight(destx,desty) or
		APBlocks(OwnerID,destx,desty) > MaxHomunRange then
		-- Check if we should stop
		if moving then
			-- Move to ourself to stop
			Move(HomunID,x,y)
		end
		
		-- Move on to the next sequence
		return true
	end

	-- Check if we're might be getting close to an intermediate location
	local nextmove = false
	if moving and seq.shortdestx ~= nil then
		-- Check our distance from the intermediate spot
		if PPDistance(x,y,seq.shortdestx,seq.shortdesty) < 3 then
			-- We're close enough, do the next move
			nextmove = true
		end
	end

	-- Check if we're currently moving
	if not moving or nextmove then
		-- We aren't moving
		
		-- Make sure we aren't just trying to get the next movement position
		if not nextmove then
			-- Get the ticks
			local ticks = GetTick()

			-- Check if we were standing still before
			if seq.oldx ~= nil then
				-- Check if we've arrived
				if PPDistance(x,y,destx,desty) < 2 then
					-- We've arrived!

					-- Log it
					Log(seq.loglevel,"Movement completed")

					return true

					--[[ Note: this is inside the check for
						oldx and oldy so that commands
						to move 1 square away are not
						ignored
					--]]
				end

				-- Check if we got stuck
				if PPDistance(seq.oldx,seq.oldy,x,y) < 2 and
					not PosLag.tries.trying then
					-- We got stuck, go to next sequence

					-- Log it
					Log(seq.loglevel,"Stuck, discarding movement command.")

					return true
				end
			else
				-- First time trying this move, log it
				Log(seq.loglevel,"Moving to position: (%d,%d)",destx,desty)
			end
		end

		-- Save the coordinates we're at
		seq.oldx,seq.oldy = x,y

		-- Make sure our distance isn't too far
		local shortened = false
		while PPDistance(x,y,destx,desty) > 12 do
			-- Make a shorter move
			destx = x-math.floor((x-destx)/2)
			desty = y-math.floor((y-desty)/2)

			-- We shortened our move
			shortened = true
		end

		-- Check if our move was shortened
		if shortened then
			-- Remember the shortened move location
			seq.shortdestx = destx
			seq.shortdesty = desty
		else
			-- Remember nothing
			seq.shortdestx = nil
		end

		-- Move towards our destination
		Move(HomunID,destx,desty)
	end

	-- Put the sequence back on the AI_seq stack
	return false,seq
end

-- Move to a spot near an actor
function SEQ_AMove(seq)
	-- Get our target's location
	local destx,desty = GetV(V_POSITION,seq[2])

	-- Check if its a valid actor
	if destx == -1 then
		return true
	end
	
	-- Check the distance parameter
	if type(seq[3]) ~= "number" then
		-- Default to 1
		seq[3] = 1
	end

	-- Make sure the distance is valid
	if math.abs(seq[3]) < 1 then
		-- Default to 1
		seq[3] = 1
	end

	-- Check if the distance is positive or negative
	if seq[3] >= 0 then
		-- Its positive

		-- Use the positive function
		seq[1] = SEQ_MOVE_TO_POS

		-- Remove any "hints"
		seq[4],seq[5] = nil,nil
	else
		-- Its negative

		-- Use the negative function
		seq[1] = SEQ_MOVE_TO_NEG

		-- Make sure we have hints
		--[[
		if type(seq[4]) ~= "number" or
			type(seq[5]) ~= "number" then

			-- Make up hints
			seq[4],seq[5] = GetV(V_POSITION,seq[2])
		end
		--]]

	end

	-- Begin using the old function immediately
	return true,seq

end

-- Move to a spot behind an actor
function SEQ_AMove_Pos(seq)
	-- Get our target's coordinates
	local destx,desty = GetV(V_POSITION,seq[2])

	-- Check that its a valid actor
	if destx == -1 then
		return true
	end

	-- Get our coordinates
	local x,y = GetV(V_POSITION,HomunID)

	-- Get our ticks
	local ticks = GetTick()

	-- Check if our actor is in range
	if (APBlocks(OwnerID,destx,desty) - seq[3] > MaxHomunRange and not seq.ignoreMaxRange) or
		not IsInSight(destx,desty) then
		-- Check if we're moving
		if HomunMotion == MOTION_MOVE then
			-- Stop
			Move(HomunID,x,y)
		end

		-- Log
		Log(seq.loglevel,"Actor #%d moved out of range, discarding follow command.",seq[2])


		-- Go to next sequence
		return true
	end
	
	-- Check if we need to recalculate a proper movement location
	if seq.preneardestx == destx and seq.preneardesty == desty then
		-- Use the saved values
		destx = seq.saveddestx
		desty = seq.saveddesty
	else
		-- Need to calculate new values

		-- Save the original values
		seq.preneardestx = destx
		seq.preneardesty = desty

		-- Get our distance
		local dist = PPDistance(x,y,destx,desty)

		-- Get the ratio of our distance, to the distance we need to move
		local distratio = (dist - seq[3]) / dist

		-- Calculate new destinations
		newdestx = x-math.floor( (x-destx) * distratio )
		newdesty = y-math.floor( (y-desty) * distratio )

		-- Check to see if we'd be backtracking
		if dist >= PPDistance(x,y,newdestx,newdesty) then
			-- Our new distance would get us closer, so use it
			destx = newdestx
			desty = newdesty
		end

		-- Save the calculated new values
		seq.saveddestx = destx
		seq.saveddesty = desty
	end

	-- Check if we've arrived
	if PPDistance(x,y,destx,desty) < 1 and seq.oldx ~= nil then
		-- Check if we're moving
		if HomunMotion == MOTION_MOVE then
			-- Stop moving
			BypassMove(HomunID,destx,desty)
		end

		-- Log it
		Log(seq.loglevel,"Arrived within %d tiles of actor #%d.",seq[3],seq[2])

		-- Go to next sequence
		return true
	end

	-- Check if we're locked within a certain distance of our owner
	if MaxDistance > 0 then
		-- Check if our destination should be modified to stay close to our owner
		if APDistance(OwnerID, destx, desty) > MaxDistance then
			-- Move on to the next sequence
			return true
		end
	end

	-- Check if we're moving
	if HomunMotion ~= MOTION_MOVE then
		-- We're not moving

		-- Check if we have coordinates of us stopped elsewhere
		if seq.oldx ~= nil then
			-- Check if we got stuck in the same place before
			if PPDistance(seq.oldx,seq.oldy,x,y) < 2 then
				-- We're probably stuck :(

				-- Check if we've run out of tries to unstick ourself
				if (seq[2] ~= OwnerID and not AllowIgnoreOwner)
					and not PosLag.tries.trying then

					-- Log it
					Log(seq.loglevel,"Stuck, discarding follow command.")

					return true
				end

				-- Check if we have tries left
				if PosLag.tries.trying then
					-- Don't interfere yet
					return false,seq
				end
			end

		elseif not seq.noInitialLog then
			-- First time moving towards the actor, log it
			Log(seq.loglevel,"Following actor #%d, until we're within %d tiles of it",
				seq[2], seq[3])
		end

		-- Save the coordinates we're at
		seq.oldx,seq.oldy = x,y

	else
		-- We're moving

		-- Check if our target is still at the same spot
		if seq.olddestx == destx and seq.olddesty == desty
			and seq.calcahead ~= true then
			-- No need to send the move command again,
			-- just continue checking later
			return false,seq
		end

	end

	-- By default, we won't be calculating ahead
	seq.calcahead = false

	-- Check if our target is moving
	if GetV(V_MOTION,seq[2]) == MOTION_MOVE then
		-- Target is moving

		-- Make sure our calc-ahead timeout is set
		if seq.calctimeout == nil then
			seq.calctimeout = 0
		end

		-- Grab the current time
		local ticks = GetTick()

		-- Check if we should calculate ahead some more
		if seq.calctimeout <= ticks then
			-- Time to calculate ahead

			-- Newly calculated destination
			local newdestx=destx
			local newdesty=desty

			-- Check if we have data from where they used to be
			if seq.olddestx ~= nil then
				-- We do, calculate a bit ahead of them
				seq.calcahead = true

				-- Get the direction
				local xdelta = destx-seq.olddestx
				local ydelta = desty-seq.olddesty

				-- Move half the distance they just moved ahead
				-- of them
				newdestx=newdestx + xdelta/2
				newdesty=newdesty + ydelta/2
			end

			-- Save the targets location
			seq.olddestx=destx
			seq.olddesty=desty

			-- Use the newly calculated destination
			destx=newdestx
			desty=newdesty

			-- Set the new timeout
			seq.calctimeout = ticks + 1000
		end

	else
		-- Target isn't moving

		-- Save the coordinates our target is at
		seq.olddestx = destx
		seq.olddesty = desty
	end
	
	-- Make sure we can move the distance
	while PPDistance(x,y,destx,desty) > 12 do
		-- Make a shorter move
		destx = x-math.floor((x-destx)/2)
		desty = y-math.floor((y-desty)/2)
	end

	-- Move towards our target
	Move(HomunID,destx,desty)

	-- Continue checking
	return false,seq
end

-- Move to a spot in front of an actor
function SEQ_AMove_Neg(seq)
	-- Get our target's coordinates
	local destx,desty = GetV(V_POSITION,seq[2])
	local targx,targy = destx,desty

	-- Check that its a valid actor
	if destx == -1 then
		return true
	end

	-- Get our original destinations
	local origx,origy

	-- Find a way to get the original target location
	if seq.oldtargx == nil then
		-- Check for seq[4] and seq[5]
		if seq[4] ~= nil and seq[5] ~= nil then
			origx,origy = seq[4],seq[5]
		else
			-- Use its current location
			origx,origy = targx,targy
		end
	else
		-- Use old targx
		origx,origy = seq.oldtargx,seq.oldtargy
	end

	-- Check if our actor is in range
	if (APBlocks(OwnerID,destx,desty) > MaxHomunRange and not seq.ignoreMaxRange) or
		not IsInSight(destx,desty) then
		-- Log
		Log(seq.loglevel,"Actor moved out of sight, discarding lead command.")

		-- Go to next sequence
		return true
	end

	-- Get our coordinates
	local x,y = GetV(V_POSITION,HomunID)

	-- Check if we're stuck from a previous move attempt
	local stuck,trying = false,PosLag.tries.trying
	if seq.oldx ~= nil and HomunMotion ~= MOTION_MOVE then
		-- Check if we got stuck in the same place before
		if PPDistance(seq.oldx,seq.oldy,x,y) < 2 then
			-- We're probably stuck :(

			stuck = true
		end

	elseif seq.olddestx == nil and not seq.noInitialLog then
		-- First time moving towards the actor, log it
		Log(seq.loglevel,"Following actor #%d, until we're %d tiles ahead of it",
			seq[2], -seq[3])
	end

	-- Check if we're "superStuck"
	local superStuck = false
	if stuck and not trying then
		superStuck = true
	end

	-- Check if we need to recalculate a proper movement location
	if seq.preneardestx == destx and seq.preneardesty == desty and not superStuck then
		-- Use the saved values
		destx = seq.saveddestx
		desty = seq.saveddesty
	else
		-- Need to calculate new values

		-- Save the original values
		local preneardestx,preneardesty = destx,desty

		-- Get the distance that they traveled
		local dist = PPDistance(origx,origy,destx,desty)

		-- Check that distance
		if dist == 0 then
			-- They haven't moved, so lets find a way to simulate their move
			if PPDistance(x,y,destx,desty) > 0 then
				-- Use the difference of homu's location and target's
				destx = targx - (targx - x)
				desty = targy - (targy - y)
			else
				-- Use a random location
				destx = targx + math.random(0,1)
				if destx == targx then
					destx = destx - 1
				end
				desty = targy + math.random(0,1)
				if desty == targy then
					desty = desty - 1
				end
			end
		end

		-- Get the new distance
		dist = PPDistance(origx,origy,destx,desty)

		-- Get the ratio of their distance
		local distratio = (dist-seq[3]) / dist

		-- Calculate a new destination
		destx = origx-math.ceil( (origx-destx) * distratio )
		desty = origy-math.ceil( (origy-desty) * distratio )

		-- Get the max range we should kite to
		local kitedistance = MaxHomunRange - 2
		if seq.kitedistance and seq.kitedistance < kitedistance then
			kitedistance = seq.kitedistance
		end

		-- Check if we should kite around someone else
		local kitearound = OwnerID
		if seq.kitearound then
			kitearound = seq.kitearound
		end

		-- Check to make sure the destination is on the screen (and we're not stuck)
		if seq[2] ~= OwnerID and seq[2] ~= kitearound and (superStuck or
			APBlocks(kitearound,destx,desty) > kitedistance or
			not IsInSight(destx,desty)) then

			-- Check if our last move failed
			local angle,radius
			if not seq.lastangle then
				-- Get the angle and radius of our move
				angle,radius = PPGetAngle(targx,targy,destx,desty)
			else
				-- Set the radius to the distance we want to be
				radius = -seq[3]

				-- Get the angle our last move
				angle = seq.lastangle
			end

			-- Owner->Homunculus angle and radius
			local o_h_angle,o_h_radius = APGetAngle(kitearound,x,y)

			-- Owner->Target angle and radius
			local o_t_angle,o_t_radius = APGetAngle(kitearound,targx,targy)

			-- Setup to loop around
			local startAngle,found,around,inc = angle,false,0,15

			-- Decide if we should go clockwise or counter clockwise
			if AngleCompare(o_h_angle,o_t_angle,180) then
				-- Since going counter-clockwise would bring us closer
				-- to our target, go clockwise instead
				inc = -inc
			end

			-- Make sure theres at least one spot we can move to
			local t_o_angle,t_o_radius = PAGetAngle(targx,targy,kitearound)
			local t_o_x,t_o_y = PPlotCircle(targx,targy,t_o_angle,radius)
			if APBlocks(kitearound,t_o_x,t_o_y) >= kitedistance or
				not IsInSight(t_o_x,t_o_y) then
				-- Log it
				Log(seq.loglevel, "No valid place to lead actor #%d to, discarding.")

				-- Drop it
				return true
			end

			-- Loop around until we find a spot that is on screen
			while not found do
				-- Increment the angle
				angle = angle + inc

				-- Make sure we're in bounds
				if angle > 360 then
					angle = angle - 360
					around = around + 1
				elseif angle < 0 then
					angle = angle + 360
					around = around + 1
				end

				-- Check if we're at our start angle again
				if (AngleCompare(angle,startAngle,inc * -2) and
					around == 1) or around > 1 then

					-- Didn't find any place
					break
				end

				-- Calculate new destx and desty
				destx,desty = PPlotCircle(targx,targy,angle,radius)

				-- Check if this new one is in range
				local blocks = APBlocks(kitearound,destx,desty)
				if (blocks <= kitedistance and not seq.ignoreMaxRange) and
					IsInSight(destx,desty) then

					-- We found a working one
					found = true
				end
			end

			-- Check if we coudn't find a place to run to
			if not found then
				-- Log it
				Log(seq.loglevel,"Couldn't find location to lead actor #%d to," ..
					" dropping lead command.",seq[2])

				-- Return
				return true
			end

			if superStuck then
				-- Log it
				Log(seq.loglevel,"Stuck against a wall while moving, trying new angle of %d.",
					angle)

				-- Save the last angle
				seq.lastangle = angle
			else
				-- Remove the last angle
				seq.lastangle = nil
			end
		end

		-- Save the before values
		seq.preneardestx = preneardestx
		seq.preneardesty = preneardesty

		-- Save the calculated new values
		seq.saveddestx = destx
		seq.saveddesty = desty
	end

	-- Check if we've arrived
	if PPDistance(x,y,destx,desty) <= 1 and seq.oldx ~= nil then
		-- Check if we're moving
		if HomunMotion == MOTION_MOVE then
			-- Stop moving
			BypassMove(HomunID,destx,desty)
		end

		-- Log it
		Log(seq.loglevel,"Arrived %d tiles ahead of actor #%d.",-seq[3],seq[2])

		-- Go to next sequence
		return true
	end

	-- Check if we're locked within a certain distance of our owner
	if MaxDistance > 0 then
		-- Check if our destination should be modified to stay close to our owner
		if APDistance(OwnerID, destx, desty) > MaxDistance then
			-- Move on to the next sequence
			return true
		end
	end

	-- Check if we're moving
	if HomunMotion == MOTION_MOVE then
		-- We're moving

		-- Check if our target is still at the same spot
		if seq.olddestx == destx and seq.olddesty == desty then
			-- No need to send the move command again,
			-- just continue checking later
			return false,seq
		end

	end

	-- Check if we're stuck
	if stuck and not seq.lastangle and
		destx == seq.olddestx and desty == seq.olddesty then
		-- Check if we're trying to move
		if not trying then
			-- Almost never ignore moving past our owner
			if seq[2] ~= OwnerID or seq[3] == -1 then
				-- Log it
				Log(seq.loglevel,"Stuck, discarding follow command.")
	
				return true
			else
				-- Reduce the distance by 1
				seq[3] = seq[3] + 1
	
				-- Unset oldx and oldy
				seq.oldx,seq.oldy = nil,nil

				-- Redo the sequence
				return true,seq
			end
		else
			-- Wait until we're not trying anymore
			return false,seq
		end
	end

	-- Make sure we can move the distance
	while PPDistance(x,y,destx,desty) > 12 do
		-- Make a shorter move
		destx = x-math.floor((x-destx)/2)
		desty = y-math.floor((y-desty)/2)
	end

	-- Save the coordinates we're at
	seq.oldx,seq.oldy = x,y

	-- Save our old destinations
	seq.olddestx,seq.olddesty = destx,desty

	-- Check if we've moved
	if targx ~= seq.checktargx and targy ~= seq.checktargy then
		-- Save old target coordinates
		seq.oldtargx,seq.oldtargy = targx,targy
	end

	-- Save to check later with
	seq.checktargx,seq.checktargy = targx,targy

	-- Move towards our target
	Move(HomunID,destx,desty)

	-- Continue checking
	return false,seq
end

-- Attack
function SEQ_Attack(seq)
	-- Make sure we have a valid target
	if Target.ID == -1 then
		-- Log it
		Log(seq.loglevel,"Invalid target, ignoring attack command")
		
		-- Invalid target
		return true
	end
	
	-- Check if our target is dead
	if GetV(V_MOTION,Target.ID) == MOTION_DEAD then
		-- Log it
		Log(seq.loglevel,"Target killed.")

		-- Turn target lock off
		SetTarget(-1)

		return true
	end

	-- Check if we would be kill stealing
	if KillSteal(Target.ID) and not Target.Lock then
		-- Log it
		Log(seq.loglevel,"Dropping target, will not kill-steal")

		-- Check if we're moving
		if HomunMotion == MOTION_MOVE then
			-- Stop moving
			BypassMove(HomunID,GetV(V_POSITION,HomunID))
		end

		-- Set target to -1
		SetTarget(-1)

		return true
	end

	-- Check if our target is in sight range
	if not ActorIsInSight(Target.ID) then
		-- Log it
		Log(seq.loglevel,"Lost sight to target, or target respawned.")

		-- Ensure target lock is off
		SetTarget(-1)

		return true
	end

	-- Get the current time
	local ticks = GetTick()

	-- Check if we've been attacking our target for too long
	if EnableIgnoreAfterTicks and Target.StartTick +
		ActorIgnoreAfterTicks < ticks then

		-- Log it
		Log(seq.loglevel,"Attacking actor #%d for too long, ignoring.",Target.ID)

		-- Ignore the target
		AddIgnore(Target.ID)

		-- Unset the target
		SetTarget(-1)

		-- Go to next sequence
		return true
	end

	-- Get our maximum distance
	local maxdist = MaxDistance
	if maxdist == 0 or MaxDistanceTimeout <= ticks then
		-- If we're not supposed to stay within range of
		maxdist = SightRange
	end

	-- Get distance of the monster to our owner
	local ownerdist = AADistance(OwnerID,Target.ID)
	
	-- Get our range
	local range = HomunRange

	-- Check if our target should only have skills used against it
	local skillsOnly = false
	if (not MAttackAllowed(Target.ID) or ownerdist >= maxdist + range)
		and MSkillsAllowed(Target.ID) then

		range = HomunSkillRange - 2
		skillsOnly = true
	end

	-- Check if skills are out of range too
	if (
		skillsOnly and (
		(ownerdist >= maxdist + range + 2 and not IgnoreAttackSkillRange) or
		(ownerdist > IgnoredRangeFromOwner and IgnoreAttackSkillRange))
	) or (
		not skillsOnly and not MSkillsAllowed(Target.ID) and
		ownerdist >= maxdist + range
	) then

		-- Determine how long to ignore them for
		local ignoreLen = nil

		-- Check if we're stuck in a short range of our owner
		if MaxDistanceTimeout > ticks then
			-- As soon as we're allowed out, this actor will be allowed again
			ignoreLen = MaxDistanceTimeout - ticks
		end

		-- Log it
		Log(seq.loglevel,"Actor #%d out of range, ignoring.",Target.ID)

		-- Add them to the ignore list
		AddIgnore(Target.ID,ignoreLen)

		-- Unset them as a target
		SetTarget(-1)

		-- Don't continue to attack
		return true
	end

	-- Get our coordinates
	local x,y = GetV(V_POSITION,HomunID)

	-- Check if our target is in range of us
	--local dist = APDistance(Target.ID,x,y) + 1
	local dist = APBlocks(Target.ID,x,y) + 1
	if dist <= range then
		-- We're close enough, check if we're moving
		if HomunMotion == MOTION_MOVE then
			-- Stop moving
			BypassMove(HomunID,x,y)
		else
			-- Stop trying
			PosLag.tries.trying = false
			PosLag.tries.left = 0
			PosLag.tries.timeout = 0
		end

		-- Check if we're using only skills
		if not skillsOnly then
			-- Trim the attack log
			local failedAttacks,timeWaited = TrimAttackLog()

			-- Check if we've waited long enough, and through
			--	enough attacks
			if timeWaited > AttacksTimeout and
				failedAttacks > AttacksMaxTries then

				-- Log it
				Log(seq.loglevel,
					"Due to %d failed attacks (waiting over %d milliseconds)," ..
					" moving to a random spot nearby.",
					failedAttacks, timeWaited)

				-- Make sure we have a jiggles count
				if seq.jiggles == nil then
					seq.jiggles = 0
				end

				-- Increment the number of jiggles
				seq.jiggles = seq.jiggles + 1

				-- Check if we've jiggled a bunch
				if seq.jiggles > 3 then
					-- Undo any target lock
					Target.Lock = false
				end

				-- Reset the tries list
				ResetAttackLog()

				-- Move to a nearby position before we try again
				return true,
					{ SEQ_MOVE, x+math.random(-1,1), y+math.random(-1,1),
						loglevel = seq.loglevel + 3 },seq
			end

			-- Set the target lock, so we don't start attacking
			-- something else
			Target.Lock = true
		
			-- Signal to attack
			SendAttack(Target.ID)

			-- Add to the attack log
			AddAttackLog()
		else
			-- Signal to send a skill, and lock if it worked
			if SendAttackSkill(Target.ID) and LockOnAttackSkill then
				Target.Lock = true
			end
		end

		-- Remove failed coords
		seq.failx = nil
	else
		-- See if using a skill against them will work
		if SendAttackSkill(Target.ID) and LockOnAttackSkill then
			Target.Lock = true
		end

		-- If we're not close enough, don't try to track attacks
		ResetAttackLog()

		--[[

The next part deals with a special sub-sequence to move towards the target.
This is used as a subsequence rather than a direct sequencer command in
the case that the target should change while chasing after it.

		--]]

		-- Check if we have a sub-sequence to chase them
		if seq.AMove_seq == nil or seq.AMove_seq[2] ~= Target.ID then
			local destx,desty = GetV(V_POSITION,Target.ID)

			-- Make one
			seq.AMove_seq = { SEQ_MOVE_TO_POS, Target.ID, 1,
				loglevel = seq.loglevel + 3, ignoreMaxRange = true,
				noInitialLog = true,
				olddestx=destx, olddesty=desty }
			seq.failx = nil
		end

		-- Use our sub sequence to get closer
		local trylater, preseq = false, seq.AMove_seq
		local tempseq
		repeat
			trylater,seq.AMove_seq = SEQ_AMove_Pos(seq.AMove_seq)
		until trylater == false or
			seq.AMove_seq == nil or
			seq.AMove_seq[1] ~= SEQ_MOVE_TO_POS

		-- Check if it worked
		if trylater then
			-- Use our old sequence again (it got cleared)
			seq.AMove_seq = preseq

			-- Check if we failed before
			if seq.failx ~= nil and HomunMotion ~= MOTION_MOVE then
				-- Check if we're at the same place
				if PPDistance(seq.failx,seq.faily,x,y) < 2 then
					-- Log it
					Log(seq.loglevel,"Couldn't get closer to our target, ignoring attack command")

					-- Add them to the ignore list
					AddIgnore(Target.ID)

					-- Remove any target lock
					SetTarget(-1)

					-- Don't continue to attack
					return true
				end
			end

			-- Save our coords
			seq.failx,seq.faily=x,y
		else
			-- Remove saved coords, we didn't fail
			seq.failx=nil

			-- Reset the movement subcommand as well
			seq.AMove_seq = nil
		end
	end

	-- Continue to try to attack
	return false,seq
end

-- Move along a set path, relative to an actor
function SEQ_PreRouted(seq)
	-- Get our target's coords
	local targx,targy = GetV(V_POSITION,seq[2])

	-- Get the current ticks
	local ticks = GetTick()

	-- Check our target
	if targx == -1 then
		-- Invalid target, so ignore this command
		return true
	end

	-- Get our coordinates
	local x,y = GetV(V_POSITION,HomunID)

	-- Get our ticks
	local ticks = GetTick()

	-- Check if our actor is in sight and range
	if not IsInSight(targx,targy) or
		APBlocks(OwnerID,targx,targy) > MaxHomunRange then
		-- Check if we're moving
		if HomunMotion == MOTION_MOVE then
			-- Stop
			BypassMove(HomunID,x,y)
		end

		-- Log
		Log(seq.loglevel,"Actor #%d moved out of range, discarding prerouted command.",
			seq[2])

		-- Go to next sequence
		return true
	end

	-- Check if we've verified our options before
	if not seq.verified then
		-- Target
		if seq[2] == HomunID then
			-- We can never make movements around ourself
			return true
		end

		-- Check our table of deltas
		if type(seq[3]) ~= "table" then
			-- Invalid parameters
			return true
		end

		-- Check each coordinate table (and find the closest to
		-- our current location
		local i,max=1,Table.getn(seq[3])
		local closeDist,closeIdx = 100,-1
		for i=1,max do
			-- Check that its a table
			if type(seq[3][i]) ~= "table" then
				-- Remove it
				max = max - 1
				Table.remove(seq[3],i)
			else
				-- Check if its a "bounce" step
				if seq[3][i].bounce then
					-- Make sure it won't get tossed
					seq[3][i][1] = 0
					seq[3][i][2] = 0
				end

				-- Verify its contents
				if type(seq[3][i][1]) ~= "number" or
					type(seq[3][i][2]) ~= "number" then

					-- Remove it
					max = max - 1
					Table.remove(seq[3],i)
				else
					-- Get the distance
					local dist = PPDistance(targx+seq[3][i][1],
								targy+seq[3][i][2],
								x,y)

					-- Check if its shorter, and allowable as a start point
					if closeDist > dist and not seq[3][i].bounce then
						-- It is, so set the new shortest
						closeDist = dist
						closeIdx = i
					end

					-- Check the distance until next move
					if type(seq[3][i][3]) ~= "number" or
						seq[3][i][3] < 0 then

						-- Set it to 1
						seq[3][i][3] = 1
					end
				end
			end
		end

		-- Verify that we have a sequence of moves
		if Table.getn(seq[3]) < 2 then
			-- We don't, so don't do anything
			return true
		end

		-- Verify that we have a beginning
		if closeIdx == -1 then
			-- No beginning, so we can't start this
			return true
		end

		-- Rotations
		if type(seq[4]) ~= "number" then
			seq[4] = 1
		end

		-- Reverse ?
		if type(seq[5]) ~= "boolean" then
			seq[5] = false
		end

		-- Set the starting step and the starting direction
		seq.startStep = closeIdx
		seq.startDir = seq[5]

		-- Initialize to 0 rotations completed (will be set to 0 immediately)
		seq.rotations = -1

		-- No failed moves yet
		seq.failedMoves = 0

		-- Make sure we don't bug out
		seq.lastMove = 0

		-- Set the first step to be the one right before our starting step
		if seq[5] then
			-- Going in reverse
			seq.step = seq.startStep + 1
		else
			-- Going forward
			seq.step = seq.startStep - 1
		end

		-- Create a function to increment the step we're on
		seq.incStep = function(seq,inc)
			-- Check for inc
			if not inc then
				inc = 1
			end

			local i
			for i=1,inc do
				-- Increment/decrement the step we're on
				if not seq[5] then
					-- Increment it
					seq.step = seq.step + 1
		
					-- Check if it should be wrapped around
					if seq.step > Table.getn(seq[3]) then
						-- Bump it back around
						seq.step = seq.step - Table.getn(seq[3])
					end
				else
					-- Decrement the step counter
					seq.step = seq.step - 1
		
					-- Check if its below 0
					if seq.step < 1 then
						-- Bump it back around
						seq.step = seq.step + Table.getn(seq[3])
					end
				end
	
				-- Check if we're at our start step again
				if seq.step == seq.startStep and seq[5] == seq.startDir then
					-- Increment the number of rotations we've done
					seq.rotations = seq.rotations + 1
				end
			end
		end

		-- Check if we have limited rotations
		local rotStr = "unlimited"
		if seq[4] > 0 then
			rotStr = tostring(seq[4])
		end

		-- Log it
		Log(seq.loglevel,"Following prerouted command (named %q) around actor #%d, %s times.",
			tostring(seq[3].name),seq[2],rotStr)

		-- Don't verify options again
		seq.verified = true

	end

	-- Check if we have other things to do while moving endlessly
	-- NOTE: While this sequence is running, its not on the sequencer queue
	if List.size(AI_seq) > 0 and (seq[4] < 1 or seq.allowBypass) then
		-- Log it
		Log(seq.loglevel,"Bypassing prerouted command (named %q) around actor #%d.",
			tostring(seq[3].name),seq[2])

		-- Allow the sequencer to do those things
		return true
	end

	-- Default to being ready for the next move
	local nextMove = true

	-- Check if we have a saved position for our target
	if seq.targx ~= nil then
		-- Check if they've moved
		if seq.targx ~= targx or seq.targy ~= targy then
			-- They've moved

			-- Check if our target is our owner
			if seq[2] == OwnerID then
				-- Ignore the rest of the prerouted command so we
				--	can use a more intelligent follow

				-- Log it
				Log(seq.loglevel,"Owner moving, so ignoring prerouted command (named %q).",
					tostring(seq[3].name))

				return true
			else
				-- Send a fresh move command
				nextMove = false
			end

		-- Otherwise check if we're not nearing our destination
		elseif APDistance(HomunID,
			seq.targx+seq[3][seq.step][1],
			seq.targy+seq[3][seq.step][2]) >=
			seq[3][seq.step][3] then

			-- Make sure we're moving (or sent a move recently)
			if HomunMotion == MOTION_MOVE then
				-- We're moving, so don't make the next move
				nextMove = false
			elseif seq.lastMove + 300 > ticks then
				-- Increment the number of times a move failed
				seq.failedMoves = seq.failedMoves + 1

				-- Alternate between trying again, and trying the next move
				if math.mod(seq.failedMoves,2) == 0 then
					-- Try again
					nextMove = false
					BypassMove(HomunID,
						seq.targx+seq[3][seq.step][1],
						seq.targy+seq[3][seq.step][2])
					seq.lastMove = ticks

				--[[
				else
					-- Try the next move
				--]]
				end
			end
		else
			-- Reset our failed moves counter
			seq.failedMoves = 0
		end
	end

	-- Check if we're supposed to move to the next spot
	if nextMove then
		seq.incStep(seq)

		-- Check if we should "bounce"
		if seq[3][seq.step].bounce then
			-- Reverse the way we're going
			seq[5] = not seq[5]

			-- Increment our step twice
			seq.incStep(seq,2)
		end
	end

	-- Generate the destination
	local destx,desty
	destx = targx + seq[3][seq.step][1]
	desty = targy + seq[3][seq.step][2]

	-- Default to waiting
	local ret = false

	-- Check if its a different destination
	if destx ~= seq.destx or desty ~= seq.desty then
		-- Save our target's coordinates
		seq.targx,seq.targy = targx,targy

		-- Save the destination coordinates
		seq.destx,seq.desty = destx,desty

		-- Check if the destination is inside our range
		if APBlocks(OwnerID,destx,desty) <= MaxHomunRange then
			-- Make the move to the destination (no anti-PosLag)
			BypassMove(HomunID,destx,desty)

			-- Set the time we last moved
			seq.lastMove = ticks
		else
			-- If we don't actually make a move, try a new one
			ret = true
		end
	end

	-- Check if we're done
	if seq[4] > 0 and seq.rotations >= seq[4] then
		-- We're done

		-- Log it
		Log(seq.loglevel,"Finished prerouted command (named %q) for actor #%d.",
			tostring(seq[3].name),seq[2])

		-- We're done now
		return true
	end

	-- Wait until the next
	return ret,seq
end

-- "Kite" a monster around
function SEQ_Kite(seq)
	-- Make sure we have a valid target
	if Target.ID == -1 then
		-- Log it
		Log(seq.loglevel,"Invalid target, ignoring kite command")
		
		-- Invalid target
		return true
	end
	
	-- Check if our target is dead
	if GetV(V_MOTION,Target.ID) == MOTION_DEAD then
		-- Log it
		Log(seq.loglevel,"Kite target died.")

		-- Turn target lock off
		SetTarget(-1)

		return true
	end

	-- Check if we would be kill stealing
	if KillSteal(Target.ID) and not Target.Lock then
		-- Log it
		Log(seq.loglevel,"Dropping target, will not kill-steal")

		-- Check if we're moving
		if HomunMotion == MOTION_MOVE then
			-- Stop moving
			BypassMove(HomunID,GetV(V_POSITION,HomunID))
		end

		-- Set target to -1
		SetTarget(-1)

		return true
	end

	-- Check if our target is in sight range
	if not ActorIsInSight(Target.ID) or 
		AABlocks(OwnerID,Target.ID) > MaxHomunRange then
		-- Log it
		Log(seq.loglevel,"Lost sight to target, or target respawned.")

		-- Ensure target lock is off
		SetTarget(-1)

		return true
	end

	-- Get the current time
	local ticks = GetTick()

	-- Check the target of our target
	if GetV(V_TARGET,Target.ID) ~= HomunID or not Target.Lock then
		-- Move closer and attack it

		-- Check if we've been at this state before
		if seq.state ~= 1 then
			-- Log it
			Log(seq.loglevel,"Trying to grab the attention of actor #%d before kite-tanking it.",
				Target.ID)

			-- Set our state
			seq.state = 1

			-- Remove the lead command
			seq.AMove_neg = nil
			seq.nlastx = nil
		end

		-- Check if we're within range
		if AADistance(HomunID,Target.ID) <= HomunRange then
			if not seq.noskills then
				-- Attack it (maybe skills)
				SendAttack()
			else
				-- Only attack
				Attack(HomunID,Target.ID)
			end

			-- Lock on
			Target.Lock = true

			-- Wait until next cycle
			return false,seq
		end

		-- Not close enough, so move closer
		-- Check if we have a sub-sequence to chase them
		if seq.AMove_pos == nil or seq.AMove_pos[2] ~= Target.ID then
			local destx,desty = GetV(V_POSITION,Target.ID)

			-- Make one
			seq.AMove_pos = { SEQ_MOVE_TO_POS, Target.ID, HomunRange - 1,
				loglevel = seq.loglevel + 3,
				noInitialLog = true,
				olddestx=destx, olddesty=desty }
			seq.failx = nil
		end

		-- Use our sub sequence to get closer
		local trylater, preseq = false, seq.AMove_pos

		repeat
			trylater,seq.AMove_pos = SEQ_AMove_Pos(seq.AMove_pos)
		until trylater == false or
			seq.AMove_pos == nil or
			seq.AMove_pos[1] ~= SEQ_MOVE_TO_POS

		if trylater then
			-- Use our old sequence again (it got cleared)
			seq.AMove_pos = preseq

			-- Get our current location
			local x,y = GetV(V_POSITION,HomunID)

			-- Check if we failed before
			if seq.failx ~= nil and HomunMotion ~= MOTION_MOVE then
				-- Check if we're at the same place
				if PPDistance(seq.failx,seq.faily,x,y) < 2 then
					-- Log it
					Log(seq.loglevel,"Couldn't get closer to our target, ignoring kite command")

					-- Add them to the ignore list
					AddIgnore(Target.ID)

					-- Remove any target lock
					SetTarget(-1)

					-- Don't continue to kite
					return true
				end
			end

			-- Set failx and faily
			seq.failx,seq.faily = x,y
		else
			-- Remove failx
			seq.failx = nil

			-- Remove subsequence
			seq.AMove_pos = nil
		end

	else
		-- Run away from it

		-- Check if we've been at this state before
		if seq.state ~= 2 then
			-- Log it
			Log(seq.loglevel,"Grabbed actor #%d's attention, kite-tanking it.",
				Target.ID)

			-- Set our state
			seq.state = 2

			-- Remove the follow subsequence
			seq.AMove_pos = nil
		end

		-- Check if we should use an attack skill
		if not seq.noskills and (IgnoreAttackSkillRange or
			AADistance(HomunID,Target.ID) <= HomunSkillRange) then

			-- Try to use one
			SendAttackSkill()
		end

		-- Check seq[3]
		if not seq[3] or seq[3] < 2 then
			seq[3] = KiteDefaultDistance
		end

		-- See if they are far enough away or they're attacking
		if AABlocks(HomunID,Target.ID) <= seq[3] or
			GetV(V_MOTION,Target.ID) == MOTION_ATTACK or
			GetV(V_MOTION,Target.ID) == MOTION_ATTACK2 then
			-- They aren't, so move

			-- Check to see if we have a subsequencer command
			if seq.AMove_neg == nil or seq.AMove_neg[2] ~= Target.ID then
				-- Make a new one
				seq.AMove_neg = { SEQ_MOVE_TO_NEG, Target.ID, -seq[3],
					noInitialLog = true,
					loglevel = seq.loglevel + 3,

					-- TODO: Allow these to be changed
					kitearound = OwnerID,
					kitedistance = MaxHomunRange - 2
				}

				-- Set the number of move tries to 0
				seq.AMove_neg_tries = 0
			end

			-- Call the sequence function
			local trylater,preseq = false,seq.AMove_neg

			repeat
				trylater,seq.AMove_neg = SEQ_AMove_Neg(seq.AMove_neg)
			until trylater == false or
				seq.AMove_neg == nil or
				seq.AMove_neg[1] ~= SEQ_MOVE_TO_NEG

			-- Check
			if seq.AMove_neg == nil or seq.AMove_neg[1] ~= SEQ_MOVE_TO_NEG then
				-- Reset the subsequence (if it got cleared)
				seq.AMove_neg = preseq
			end

			-- Get our location
			local x,y = GetV(V_POSITION,HomunID)

			-- Check if we've run before
			local removeSeq = true
			if seq.nlastx ~= nil and HomunMotion ~= MOTION_MOVE then
				-- Check if we're at the same place
				if PPDistance(seq.nlastx,seq.nlasty,x,y) < 1 then
					-- Check if we've tried too many times
					if seq.AMove_neg_tries >= 4 then
						-- Log it
						Log(seq.loglevel,"Couldn't get farther from our target," ..
							" kiting to our owner.")


						-- Check how far we are from our owner
						local dist = HomunFollowRange
						if APDistance(OwnerID,x,y) <= dist then
							-- Reverse it, so we go the other way
							dist = -dist
						end
	
						-- Continue to kite after running to owner
						return true,{ SEQ_MOVE_TO, OwnerID, dist },seq
					end

					-- Increment the number of tries
					seq.AMove_neg_tries = seq.AMove_neg_tries + 1

					-- Keep the sequence
					removeSeq = false
				else
					-- Set nlastx and nlasty
					seq.nlastx,seq.nlasty = x,y
				end
			else
				-- Set nlastx and nlasty
				seq.nlastx,seq.nlasty = x,y
			end

			-- Check if we should reuse the sequence
			if (seq.AMove_neg == nil or seq.AMove_neg[1] ~= SEQ_MOVE_TO_NEG) and
				not removeSeq then
				-- Reuse it
				seq.AMove_neg = preseq
			else
				-- Make a fresh one next time around
				seq.AMove_neg = nil
			end
		end
	end

	-- Wait until the next cycle
	return false,seq
end

-- Use a skill on an actor
function SEQ_ASkill(seq)

--[[ Note:
	This sequencer command is only used when the user manually
	tells us to use a skill
--]]

	-- Get the range
	local range --= GetV(V_SKILLATTACKRANGE,HomunID,seq[3])
	if AttackSkillsByID[seq[3]] then
		range = AttackSkillsByID[seq[3]].range
	else
		range = GetV(V_SKILLATTACKRANGE,HomunID,seq[3])
	end

	-- Check if the target is too far away
	if (AADistance(seq[4],HomunID) > range and not IgnoreAttackSkillRange) or
		(AADistance(seq[4],OwnerID) > IgnoredRangeFromOwner and IgnoreAttackSkillRange) then

		-- Check if the target is too far away for us to cast
		if AADistance(seq[4],OwnerID) > IgnoredRangeFromOwner then
			-- Log it
			Log(seq.loglevel,
				"Actor #%d is too far away to cast skill (ID: %d, level %d).",
					seq[4],seq[3],seq[2])

			-- Ignore the command
			return true
		end

		-- Check if the target is in sight range of our owner
		if not ActorIsInSight(seq[4]) then
			-- Log it
			Log(seq.loglevel,
				"Lost sight of manual skill (ID: %d, level %d) target #%d",
				seq[3], seq[2], seq[4])

			-- Ignore this command
			return true
		end

		-- Get the range to move to the target
		local targetrange = range - 2

		-- Make sure its a sane target
		if targetrange < 1 then
			targetrange = 1
		end

		-- We need to get closer to the target, so queue
		-- another sequence ahead of this one (and allow
		-- the sequencer to process it right away)
		return true,{ SEQ_MOVE_TO, seq[4], targetrange, loglevel=seq.loglevel + 3 },seq
	end

	-- Log it
	Log(seq.loglevel,"Casting manual skill (ID: %d, level %d) at Actor #%d",
		seq[3],seq[2],seq[4])

	-- Send our skill-use command
	SkillObject(HomunID,seq[2],seq[3],seq[4])

	-- Allow the sequencer to process the next command
	return true
end

-- Use a skill on a position
function SEQ_PSkill(seq)

--[[ Note:
	This sequencer command is only used when the user manually
	tells us to use a skill
--]]

	-- Check if the target is too far away
	local range = GetV(V_SKILLATTACKRANGE,HomunID,seq[3])
	if APDistance(HomunID,seq[4],seq[5]) > range then

		-- Check if the target is in sight range of our owner
		if not IsInSight(seq[4],seq[5]) then
			-- Log it
			Log(seq.loglevel,
				"Lost sight of manual skill (ID: %d, level %d) target area (%d,%d)",
				seq[3], seq[2], seq[4], seq[5])

			-- Ignore this command
			return true
		end

		-- We need to get closer to the target, so queue
		-- another sequence ahead of this one (and allow
		-- the sequencer to process it right away)
		return true,{ SEQ_MOVE, seq[4], seq[5], range-2, loglevel=seq.loglevel + 3 },seq
	end

	-- Log it
	Log(seq.loglevel,
		"Casting manual skill (ID: %d, level %d) on location (%d,%d)",
		seq[3],seq[2],seq[4],seq[5])

	-- Send our skill-use command
	SkillGround(HomunID,seq[2],seq[3],seq[4],seq[5])

	-- Allow the sequencer to process the next command
	return true
end
