--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

------------------------
-- Distance Functions --
------------------------

-- Position->Position Pythagorean Distance
function PPDistance(x1,y1,x2,y2)
	-- Check all our parameters
	if type(x1) ~= "number" or
		type(y1) ~= "number" or
		type(x2) ~= "number" or
		type(y2) ~= "number"
	then
		-- Something wasn't a number, return nil so its easier to trace
		return nil
	end

	--return math.floor(math.sqrt( (x2-x1)^2 + (y2-y1)^2 ))
	return math.sqrt( (x2-x1)^2 + (y2-y1)^2 )
end

-- Actor->Position Pythagorean Distance
function APDistance(actor,x,y)
	-- Check actor
	if type(actor) ~= "number" then
		-- Return nil so its easier to trace
		return nil
	end

	local x2,y2 = GetV(V_POSITION,actor)

	if x2 ~= -1 then
		return PPDistance(x2,y2,x,y)
	end

	-- If there was an invalid actor, return a high number
	return SightRange * SightRange
end

-- Actor->Actor Pythagorean Distance
function AADistance(actor1,actor2)
	-- Check actor2
	if type(actor2) ~= "number" then
		-- Return nil so its easier to trace
		return nil
	end

	local x,y = GetV(V_POSITION,actor2)

	if x ~= -1 then
		return APDistance(actor1,x,y)
	end

	-- If there was an invalid actor, return a high number
	return SightRange * SightRange
end

-- Position->Position Block Distance
function PPBlocks(x1,y1,x2,y2)
	local xdist = math.abs(x2-x1)
	local ydist = math.abs(y2-y1)

	if xdist > ydist then
		return xdist
	else
		return ydist
	end
end

-- Actor->Position Block Distance
function APBlocks(actor,x,y)
	-- Check actor
	if type(actor) ~= "number" then
		-- Return nil so its easier to trace
		return nil
	end

	local x2,y2 = GetV(V_POSITION,actor)

	if x2 ~= -1 then
		return PPBlocks(x2,y2,x,y)
	end

	-- If there was an invalid actor, return a high number
	return SightRange * SightRange
end

-- Actor->Actor Block Distance
function AABlocks(actor1,actor2)
	-- Check actor 2
	if type(actor2) ~= "number" then
		-- Return nil so its easier to trace
		return nil
	end

	local x,y = GetV(V_POSITION,actor2)

	if x ~= -1 then
		return APBlocks(actor1,x,y)
	end

	-- If there was an invalid actor, return a high number
	return SightRange * SightRange
end

-- Find actor nearest to a position from a table
function PTNearest(x,y,t)
	-- Check the size of the table
	if Table.getn(t) < 1 then
		return -1
	end

	-- Loop through the table
	local i,dist=0,SightRange * SightRange
	local nearest=dist
	local index=-1
	for i=1,t.n do
		-- Make sure its an actor number
		if type(t[i]) == "number" then
			-- Get the distance to the actor
			dist=APDistance(t[i],x,y)
		end

		-- Check if its closer
		if dist < nearest then
			-- Set the new nearest actor
			nearest=dist
			index=i
		end
	end

	-- Return the nearest
	return t[index]
end

-- Find actor nearest to another from a table
function ATNearest(actor,t)
	-- Check the size of the table
	if Table.getn(t) < 1 then
		return -1
	end

	-- Get the actor's current position
	local x,y = GetV(V_POSITION,actor)

	-- Get the nearest from the table
	return PTNearest(x,y,t)
end

---------------------------
-- Sight-range Functions --
---------------------------

-- Check if a position is in sight-range of our owner
function IsInSight(x,y)
	if APBlocks(OwnerID,x,y) > SightRange then
		return false
	end

	return true
end

-- Check if an actor is in sight-range of our owner
function ActorIsInSight(actor)
	if AABlocks(actor,OwnerID) > SightRange then
		return false
	end

	return true
end

----------------------
-- Circle Functions --
----------------------

-- Function to get the angle from one point to another
function PPGetAngle(x1,y1,x2,y2)
	-- Get the radius
	local radius = PPDistance(x1,y1,x2,y2)

	-- Check if the radius is 0
	if radius == 0 then
		-- Return an angle of -1
		return -1,0
	end

	-- Get the changes
	local xdelta = x2 - x1
	local ydelta = y2 - y1

	-- Get the angle from X
	local angle = math.deg(math.acos(xdelta / radius))

	-- Get the angle from Y
	if math.deg(math.asin(ydelta / radius)) < 0 then
		angle = 360 - angle
	end

	-- Return the angle and the radius
	return angle,radius
end

-- Function to get the angle from a point to an actor
function PAGetAngle(x1,y1,actor)
	-- Get the actor's coordinates
	local x2,y2 = GetV(V_POSITION,actor)

	-- Check the actor's coordinates for sanity
	if x2 == -1 then
		-- Actor's position is unknown
		return -1,0
	end

	-- Reuse above code
	return PPGetAngle(x1,y1,x2,y2)
end

-- Function to get the angle from an actor to a position
function APGetAngle(actor,x2,y2)
	-- Get the actor's coordinates
	local x1,y1 = GetV(V_POSITION,actor)

	-- Check the actor's coordinates for sanity
	if x1 == -1 then
		-- Actor's position is unknown
		return -1,0
	end

	-- Reuse above code
	return PPGetAngle(x1,y1,x2,y2)
end

-- Function to get the angle from one actor to another
function AAGetAngle(actor1,actor2)
	-- Get the first actor's coordinates
	local x,y = GetV(V_POSITION,actor1)

	-- Check the actor's coordinates for sanity
	if x == -1 then
		-- Actor1's position is unknown
		return -1,0
	end

	-- Reuse above code
	return PAGetAngle(x,y,actor2)
end

-- Function to plot a point on a circle centered around a position
function PPlotCircle(x,y,angle,radius)
	-- Get the angle in radians
	local rads = math.rad(angle)

	-- Get the xdelta and ydeltas
	local xdelta = radius * math.cos(rads)
	local ydelta = radius * math.sin(rads)

	-- Return the locations
	return x + xdelta, y + ydelta
end

-- Function to plot a point on a circle centered around an actor
function APlotCircle(actor,angle,radius)
	-- Get the actor's coordinates
	local x,y = GetV(V_POSITION,actor)

	-- Check the actor's coordinates for sanity
	if x == -1 then
		-- Return -1s for the coordinates
		return -1,-1
	end

	-- Reuse above code
	PPlotCircle(x,y,angle,radius)
end

-- Check to see if an angle is within X radians of a second
function AngleCompare(angle1,angle2,radians)
	local bottom,top

	-- Get the modulus remainder of both angles
	angle1 = math.mod(angle1,360)
	angle2 = math.mod(angle2,360)

	-- Check if radians is positive or negative
	if radians > 0 then
		-- Check if adding radians to angle1 would throw it past 359
		if angle1 + radians >= 360 then
			-- Make a second check
			if AngleCompare(0,angle2,math.mod(angle1 + radians,360)) then
				-- Found it already
				return true
			end

			-- Reduce radians below 360
			radians = radians - math.mod(angle1 + radians,360) - 1
		end

		-- Set the bottom and the top
		bottom = angle1
		top = angle1 + radians

	elseif radians < 0 then
		-- Invert the radians, to make later stuff easier to read
		radians = -radians

		-- Check if subtracting radians would throw it under 0
		if angle1 - radians < 0 then
			-- Make a second check
			if AngleCompare(360 + angle1 - radians,angle2,math.abs(angle1 - radians) - 1) then
				-- Found it already
				return true
			end

			-- Increase the radians above -1
			radians = radians + math.abs(angle1 - radians)
		end

		-- Set the bottom
		bottom = angle1 - radians
		top = angle1

	else
		-- If radians is 0, then check if the angles are the same
		if angle1 == angle2 then
			-- They check out as the same
			return true
		end

		-- They're not the same
		return false
	end

	-- Check if its above the bottom but below the top, inclusively
	if bottom <= angle2 and angle2 <= top then
		-- Its inside
		return true
	end

	-- Its not inside
	return false
end

-------------------
-- Actor Support --
-------------------

-- Check if an actor is in a table
function ActorIsInTable(actor,t)
	-- Loop through the table
	local i,max = 1,Table.getn(t)
	for i=1,max do
		-- Check if its the actor
		if t[i] == actor then
			-- It is, return true
			return true
		end
	end
	
	-- Otherwise the actor isn't in the table
	return false
end

-- Identify various bits of information, such as all the attackers
-- of a specific actor, members of the party, etc
function GetActorInfo()
	-- Get all the actors on the screen
	local actors = GetActors()

	-- Clear our table of actor targets
	ActorTargets = { }

	-- Clear our table of temporary friends
	TempFriends = { }

	-- Clear our table of permanent friends that are on screen
	PermFriendsOnScr = { }

	-- Get the current time
	local ticks = GetTick()

	-- Loop through each actor, gathering data
	local i,actor,target
	for i,actor in ipairs(actors) do
		-- Get the actor's motion
		local motion = GetV(V_MOTION,actor)

		-- Check if the actor is getting damaged
		if motion == MOTION_DAMAGE then
			-- They're getting damages, so set their
			--	damaged time to right now
			ActorDamageTimes[actor] = ticks
		else
			-- Check if this actor has been seen before
			if ActorDamageTimes[actor] == nil then
				-- Set it to 0, so we won't have any comparison
				--	errors
				ActorDamageTimes[actor] = 0
			end
		end

		-- Check if the current actor has anything targetting it
		if ActorTargets[actor] == nil then
			-- Nope

			-- Create a table for it
			ActorTargets[actor] = Table.new()
		end

		-- Check if the actor is in a position to target something
		if motion ~= MOTION_DEAD and motion ~= MOTION_SIT then
			-- It is

			-- Get the current actor's target
			target = GetV(V_TARGET,actor)

			-- Check if the actor is targetting anything
			if target ~= 0 then
				-- Check if the target has a table yet
				if ActorTargets[target] == nil then
					-- Nope

					-- Create a table for it
					ActorTargets[target] = Table.new()
				end

				-- Add the actor as an attacker of the target
				Table.insert(ActorTargets[target],actor)
			end

		end

		-- Check if the actor is a permanent friend
		if IsPermFriend(actor) then
			-- Add it to our friends who're on screen
			PermFriendsOnScr[actor] = true
		end

		-- Should we gather friend data?
		if DefendFriends and actor ~= HomunID and actor ~= OwnerID then
			-- Check if the actor is a monster
			if IsMonster(actor) ~= 1 then
				-- NOTE: Homunculi show as nonmonsters in nonpvp zones

				-- Check if they're close enough to our owner
				if AABlocks(OwnerID,actor) <= TemporaryFriendTiles then
					-- They're a temporary friend
					TempFriends[actor] = true
				else
					-- They're not a temporary friend
					TempFriends[actor] = false
				end

				--[[ See if we can check their health
				-- NOTE: This idea was based on false "information"
				--		found on the RagnaInfo forums.
				if GetV(V_HP,actor) ~= -1 then
					-- Log it
					Log(0,"Identified party member as friend: Actor #%d",actor)

					-- Add them to the friend table
					AddFriend(actor)
				end
				--]]
			end
		end
	end
end

-------------------
-- HP/SP Support --
-------------------

-- Calculate the delta of our own and our owner's HPs
function CalculateHPDeltas()
	-- Get the current tick count
	local ticks = GetTick()

	-- Get our owner's HP value
	local hp = GetV(V_HP,OwnerID)

	-- Remove any HP values that are older than 1000 milliseconds
	local other
	repeat
		-- Grab the oldest HP value
		other = List.popright(OwnerHPTracking)

		-- Check the validity
		if other == nil then
			other = { ticks, hp }
		end
	until ticks - other[1] <= 1000

	-- Its new enough, push it back on
	List.pushright(OwnerHPTracking,other)

	-- Get the delta
	OwnerHPDelta.second = hp - other[2]

	-- Now grab the newest HP value
	other = OwnerHPTracking[OwnerHPTracking.first]

	-- Get the delta
	OwnerHPDelta.iteration = hp - other[2]

	-- Now add this into our log
	List.pushleft(OwnerHPTracking,{ ticks, hp })

	-- Begin doing the exact same for homunculus
	hp = GetV(V_HP,HomunID)

	-- Remove any HP values that are older than 1000 milliseconds
	repeat
		-- Grab the oldest HP value
		other = List.popright(HomunHPTracking)

		-- Check the validity
		if other == nil then
			other = { ticks, hp }
		end
	until ticks - other[1] <= 1000

	-- Its new enough, so put it back
	List.pushright(HomunHPTracking,other)

	-- Get the delta
	HomunHPDelta.second = hp - other[2]

	-- Now grab the newest HP value
	other = HomunHPTracking[HomunHPTracking.first]

	-- Get the delta
	HomunHPDelta.iteration = hp - other[2]

	-- Now add this into our hp log
	List.pushleft(HomunHPTracking,{ ticks, hp })
end

-- Determine the amount of HP remaining
function ActorHPAmount(actor)
	-- Get the actor's HP amount
	return GetV(V_HP,actor)
end

-- Determine the percentage of SP remaining
function ActorHPPercent(actor)
	-- Get the actor's HP and max HP amounts
	local hp = GetV(V_HP,actor)
	local max = GetV(V_MAXHP,actor)

	-- Turn it into a percentage and return
	return math.floor( (hp/max) * 100 )
end

-- Determine the maximum amount of SP that is usable
function SPUsableAmount(actor)
	-- Decide on the minimum amount
	local min = 0

	-- Check if we're getting our homunculus's SP or our alchemist's SP
	if actor == HomunID then
		min = MinSP
	elseif actor == OwnerID then
		min = OwnerMinSP

		-- Check if we have autores enabled
		if AutoRes then
			min = min + 50
		end
	end

	-- Subtract our minimum SP amount
	return ActorSP(actor) - min
end

-- Determine the maximum amount of SP that is usable, using percents
function SPUsablePercent(actor)
	-- Get the percent our SP is at
	local sp = ActorSPPercent(actor)

	-- Decide on the minimum amount
	local min = 0

	-- Check if we're getting our homunculus's SP or our alchemist's SP
	if actor == HomunID then
		min = MinSP
	elseif actor == OwnerID then
		min = OwnerMinSP
	end

	-- Determine the difference
	sp = sp - min

	-- Unpercentize it
	sp = math.floor((sp * GetV(V_MAXSP,actor) / 100))

	-- Check if we need to save some for autores
	if actor == OwnerID and AutoRes then
		-- Save 50 for AutoRes
		sp = sp - 50
	end

	-- Return the SP
	return sp
end

-- Determine the amount of SP remaining
function ActorSPAmount(actor)
	-- Get the actor's SP amount
	return GetV(V_SP,actor)
end

-- Determine the percentage of SP remaining
function ActorSPPercent(actor)
	-- Get the actor's SP and max SP amounts
	local sp = GetV(V_SP,actor)
	local max = GetV(V_MAXSP,actor)

	-- Turn it into a percentage and return
	return math.floor( (sp/max) * 100 )
end

-------------------------
-- Debugging Functions --
-------------------------

-- Normal log-style
function Log(level,text,...)
	-- Check the level is too high
	if level > DebugLevel then
		return
	end

	-- Format the string
	local str = "(" .. level .. ") " .. LogFormat(text,unpack(arg))

	-- Check if its the same as the last log line
	if LastLogLine == str then
		-- Add a flag to say its a duplicate
		str = str .. " (duplicate)"
	end

	-- Save the log line
	LastLogLine = str

	-- Make the actual log
	TraceAI(str)
end

-- Workaround for logs when TraceAI.txt is only 1 line
function LogWorkaround(level,text,...)
	-- Check if the level is too high
	if level > DebugLevel then
		return
	end

	-- Format the text
	local str = "(" .. level .. ") " .. LogFormat(text,unpack(arg))

	-- Put the text into the log
	List.pushright(Logs,str)

	-- See if we need to remove a line
	if List.size(Logs) > LogLines then
		List.popleft(Logs)
	end

	-- Gather all the lines into 1 string
	str = "\r\n"
	local i
	for i=Logs.first,Logs.last do
		str = str .. Logs[i] .. "\r\n"
	end

	-- Output via TraceAI
	TraceAI(str)
end

-- Function to verify that we have enough parameters to a string.format, and then format it
function LogFormat(text,...)
	-- Default to having valid parameters
	local valid = true

	-- Find all places that a % operator shows up
	local curParam,paramStart,paramEnd = 1,nil,1
	while true do
		-- Get the next match
		paramStart,paramEnd = string.find(text,"%%%d-%.?%d-[dfiqs]",paramEnd)

		-- Check that we found a place
		if paramStart == nil then
			-- We didn't, stop looping
			break
		end

		-- Get the type that we need
		local t = string.sub(text,paramEnd,paramEnd)
		if t == "q" or t == "s" then
			-- We need a string
			t = "string"
		else
			-- We need a number
			t = "number"
		end

		-- Check if our parameter is of that type
		if type(arg[curParam]) ~= t then
			-- Its not, set valid to false
			valid = false

			-- Stop looping
			break
		end

		-- Increment our current parameter
		curParam = curParam + 1
	end

	-- Check if we're valid
	if valid then
		-- Format as normal
		return string.format(text,unpack(arg))
	else
		-- Format to an error
		local str = string.format("Error with LogFormat(%q",text)

		-- Add all the arguments
		local i,form
		for i=1,Table.getn(arg) do
			-- Check if we should put quotes around it
			if type(arg[i]) == "string" then
				form = "%q"
			else
				form = "%s"
			end
			-- Add the parameter
			str = str .. string.format(","..form,tostring(arg[i]))
		end

		-- Add the end parenthesis
		str = str .. ")"

		-- Return our string
		return str
	end
end

--------------------
-- Math Functions --
--------------------

-- Return the modulus remainder of a number (unneeded -- math.mod serves the purpose)
--[[
function Modulus(a,b)
	return a - b*math.floor(a/b)
end
--]]

---------------------------
-- Function Replacements --
---------------------------

function NoReturn()
end

function ZeroReturn()
	return 0
end

function ZeroZeroReturn()
	return 0,0
end

function NegativeReturn()
	return -1
end

function TrueReturn()
	return true
end

function FalseReturn()
	return false
end

function PassThru(arg)
	return arg
end